# Copyright 2025 by Vytautas Liuolia.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Information on stable Falcon releases.

This extension aggregates a table of maintained and EOL Falcon releases from
the provided path containing changelogs.
"""

import collections
import datetime
import enum
import itertools
import pathlib
import re

import sphinx.util.docutils

FALCON_DOCS_ROOT = pathlib.Path(__file__).resolve().parent.parent


class _FalconRelease:
    """A stable Falcon release."""

    _CHANGELOG_PATTERN = re.compile(r'^(\d+)\.(\d+).(\d+)\.rst$')
    _RELEASE_META_PATTERN = re.compile(r'\.\.\s+falcon-release\:\s+([\d-]+)')

    _PYPI_API_URL = 'https://pypi.org/pypi/falcon/{version}/json'

    @classmethod
    def from_changelog(cls, path):
        if not (mt := cls._CHANGELOG_PATTERN.match(path.name)):
            return None

        version = tuple(map(int, mt.groups()))
        if mt := cls._RELEASE_META_PATTERN.search(path.read_text()):
            date = datetime.date.fromisoformat(mt.group(1))
        else:
            date = None

        return cls(version, date)

    @classmethod
    def gather(cls, changelog_path):
        releases = [cls.from_changelog(path) for path in changelog_path.iterdir()]
        return sorted(
            release
            for release in releases
            if release is not None and not release.is_zerover
        )

    @property
    def is_zerover(self):
        return self.major == 0

    @property
    def major(self):
        return self.version[0]

    @property
    def version_str(self):
        return '.'.join(map(str, self.version))

    @property
    def version_major_minor(self):
        return '.'.join(map(str, self.version[:2]))

    def __init__(self, version, date=None):
        self.version = version
        self.date = date

    @property
    def doc_link(self):
        return f':doc:`{self.version_str} </changes/{self.version_str}>`'

    @property
    def doc_link_major_minor(self):
        return f':doc:`{self.version_major_minor} </changes/{self.version_str}>`'

    @property
    def rst_repr(self):
        return f'{self.doc_link} ({self.date})'

    def __eq__(self, other):
        return self.version == other.version

    def __lt__(self, other):
        return self.version < other.version

    def __repr__(self):
        return f'<falcon-{self.version_str} {self.date}>'

    def _load_date_from_pypi(self):
        # NOTE(vytas): This is not really used, but it is handy to verify the
        #   dates of historical releases, and other types of debugging.
        import requests

        resp = requests.get(self._PYPI_API_URL.format(version=self.version_str))
        if resp.status_code == 200:
            artifact, *_ = resp.json()['urls']
            upload_date, _, _ = artifact['upload_time'].partition('T')
            self.date = datetime.date.fromisoformat(upload_date)
        return self.date


_FalconMajorMinor = collections.namedtuple('_FalconMajorMinor', ('first', 'latest'))


class _ReleaseStatus(enum.Enum):
    current = 'current'
    security = 'security'
    eol = 'eol'

    @property
    def ref(self):
        return f':ref:`stable_release_{self.value}`'


class FalconReleasesDirective(sphinx.util.docutils.SphinxDirective):
    """Directive to tabulate the summary of stable Falcon releases."""

    _OLDSTABLE_MAINTENANCE_PERIOD = datetime.timedelta(days=365 * 2)

    required_arguments = 1
    has_content = True

    def run(self):
        changelog_path = pathlib.Path(self.arguments[0])
        if not changelog_path.is_absolute():
            changelog_path = FALCON_DOCS_ROOT / changelog_path

        releases = _FalconRelease.gather(changelog_path)
        for release in releases[:-1]:
            assert release.date is not None, (
                f'{release} has no release date metadata (and is not latest)'
            )
        if releases[-1].date is None:
            releases.pop()

        stable_releases = []
        for _, group in itertools.groupby(
            reversed(releases), lambda rel: rel.version_major_minor
        ):
            series = list(group)
            stable_releases.append(
                _FalconMajorMinor(first=min(series), latest=max(series))
            )

        current = stable_releases[0]
        current_series = sorted(
            rel for rel in stable_releases if rel.first.major == current.first.major
        )
        oldstable = max(
            rel for rel in stable_releases if rel.first.major == current.first.major - 1
        )

        header = '.. list-table::\n    :header-rows: 1\n\n'
        rows = [
            (
                'Major/Minor Version',
                'First Release',
                'Latest Patch Release',
                'Release Status',
            )
        ]

        for first, latest in stable_releases:
            eol_label = ''
            if latest == current.latest:
                status = _ReleaseStatus.current
            elif latest == oldstable.latest:
                eol_date = (
                    current_series[0].first.date + self._OLDSTABLE_MAINTENANCE_PERIOD
                )
                if datetime.date.today() < eol_date:
                    status = _ReleaseStatus.security
                    # NOTE(vytas): Move to a new line because otherwise table
                    #   layout gets messed up in an unpleasant way.
                    eol_label = f'\n\n        (until {eol_date})'
                else:
                    status = _ReleaseStatus.eol
                    eol_label = f'(since {eol_date})'
            else:
                status = _ReleaseStatus.eol

            rows.append(
                (
                    first.doc_link_major_minor,
                    first.rst_repr,
                    latest.version_str if latest == first else latest.rst_repr,
                    f'{status.ref} {eol_label}',
                )
            )

        content = '\n\n'.join(
            '    * ' + '\n      '.join(f'- {cell}' for cell in row) for row in rows
        )

        return self.parse_text_to_nodes(header + content)


def setup(app):
    app.add_directive('falcon-releases', FalconReleasesDirective)

    return {
        'version': '0.1',
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



import pytest

import falcon
import falcon.errors as errors
import falcon.status_codes as status


@pytest.mark.parametrize(
    'err, title',
    [
        (falcon.HTTPBadRequest, status.HTTP_400),
        (falcon.HTTPUnauthorized, status.HTTP_401),
        (falcon.HTTPForbidden, status.HTTP_403),
        (falcon.HTTPNotFound, status.HTTP_404),
        (errors.HTTPRouteNotFound, status.HTTP_404),
        (falcon.HTTPNotAcceptable, status.HTTP_406),
        (falcon.HTTPConflict, status.HTTP_409),
        (falcon.HTTPGone, status.HTTP_410),
        (falcon.HTTPLengthRequired, status.HTTP_411),
        (falcon.HTTPPreconditionFailed, status.HTTP_412),
        (falcon.HTTPContentTooLarge, status.HTTP_413),
        (falcon.HTTPUriTooLong, status.HTTP_414),
        (falcon.HTTPUnsupportedMediaType, status.HTTP_415),
        (falcon.HTTPUnprocessableEntity, status.HTTP_422),
        (falcon.HTTPLocked, status.HTTP_423),
        (falcon.HTTPFailedDependency, status.HTTP_424),
        (falcon.HTTPPreconditionRequired, status.HTTP_428),
        (falcon.HTTPTooManyRequests, status.HTTP_429),
        (falcon.HTTPRequestHeaderFieldsTooLarge, status.HTTP_431),
        (falcon.HTTPUnavailableForLegalReasons, status.HTTP_451),
        (falcon.HTTPInternalServerError, status.HTTP_500),
        (falcon.HTTPNotImplemented, status.HTTP_501),
        (falcon.HTTPBadGateway, status.HTTP_502),
        (falcon.HTTPServiceUnavailable, status.HTTP_503),
        (falcon.HTTPGatewayTimeout, status.HTTP_504),
        (falcon.HTTPVersionNotSupported, status.HTTP_505),
        (falcon.HTTPInsufficientStorage, status.HTTP_507),
        (falcon.HTTPLoopDetected, status.HTTP_508),
        (falcon.HTTPNetworkAuthenticationRequired, status.HTTP_511),
    ],
)
def test_with_default_title_and_desc(err, title):
    with pytest.raises(err) as e:
        raise err()

    assert e.value.title == title
    assert e.value.description is None

    if e.value.headers:
        assert 'Retry-After' not in e.value.headers


@pytest.mark.parametrize(
    'err, title, args',
    (
        (falcon.HTTPMethodNotAllowed, status.HTTP_405, (['GET'],)),
        (falcon.HTTPRangeNotSatisfiable, status.HTTP_416, (11,)),
    ),
)
def test_with_default_title_and_desc_args(err, title, args):
    with pytest.raises(err) as e:
        raise err(*args)

    assert e.value.title == title
    assert e.value.description is None

    if e.value.headers:
        assert 'Retry-After' not in e.value.headers


@pytest.mark.parametrize(
    'err',
    [
        falcon.HTTPBadRequest,
        falcon.HTTPUnauthorized,
        falcon.HTTPForbidden,
        falcon.HTTPNotFound,
        errors.HTTPRouteNotFound,
        falcon.HTTPNotAcceptable,
        falcon.HTTPConflict,
        falcon.HTTPGone,
        falcon.HTTPLengthRequired,
        falcon.HTTPPreconditionFailed,
        falcon.HTTPContentTooLarge,
        falcon.HTTPUriTooLong,
        falcon.HTTPUnsupportedMediaType,
        falcon.HTTPUnprocessableEntity,
        falcon.HTTPLocked,
        falcon.HTTPFailedDependency,
        falcon.HTTPPreconditionRequired,
        falcon.HTTPTooManyRequests,
        falcon.HTTPRequestHeaderFieldsTooLarge,
        falcon.HTTPUnavailableForLegalReasons,
        falcon.HTTPInternalServerError,
        falcon.HTTPNotImplemented,
        falcon.HTTPBadGateway,
        falcon.HTTPServiceUnavailable,
        falcon.HTTPGatewayTimeout,
        falcon.HTTPVersionNotSupported,
        falcon.HTTPInsufficientStorage,
        falcon.HTTPLoopDetected,
        falcon.HTTPNetworkAuthenticationRequired,
    ],
)
def test_with_title_desc_and_headers(err):
    title = 'trace'
    desc = 'boom'
    headers = {'foo': 'bar'}

    with pytest.raises(err) as e:
        raise err(title=title, description=desc, headers=headers)

    assert e.value.title == title
    assert e.value.description == desc
    assert e.value.headers['foo'] == 'bar'


@pytest.mark.parametrize(
    'err',
    [
        falcon.HTTPBadRequest,
        falcon.HTTPUnauthorized,
        falcon.HTTPForbidden,
        falcon.HTTPNotFound,
        errors.HTTPRouteNotFound,
        falcon.HTTPNotAcceptable,
        falcon.HTTPConflict,
        falcon.HTTPGone,
        falcon.HTTPLengthRequired,
        falcon.HTTPPreconditionFailed,
        falcon.HTTPContentTooLarge,
        falcon.HTTPUriTooLong,
        falcon.HTTPUnsupportedMediaType,
        falcon.HTTPUnprocessableEntity,
        falcon.HTTPLocked,
        falcon.HTTPFailedDependency,
        falcon.HTTPPreconditionRequired,
        falcon.HTTPTooManyRequests,
        falcon.HTTPRequestHeaderFieldsTooLarge,
        falcon.HTTPUnavailableForLegalReasons,
        falcon.HTTPInternalServerError,
        falcon.HTTPNotImplemented,
        falcon.HTTPBadGateway,
        falcon.HTTPServiceUnavailable,
        falcon.HTTPGatewayTimeout,
        falcon.HTTPVersionNotSupported,
        falcon.HTTPInsufficientStorage,
        falcon.HTTPLoopDetected,
        falcon.HTTPNetworkAuthenticationRequired,
    ],
)
def test_kw_only(err):
    with pytest.raises(TypeError, match='positional argument'):
        err('foo', 'bar')


@pytest.mark.parametrize(
    'err, args',
    (
        (falcon.HTTPMethodNotAllowed, (['GET'],)),
        (falcon.HTTPRangeNotSatisfiable, (11,)),
    ),
)
def test_with_title_desc_and_headers_args(err, args):
    title = 'trace'
    desc = 'boom'
    headers = {'foo': 'bar'}

    with pytest.raises(err) as e:
        raise err(*args, title=title, description=desc, headers=headers)

    assert e.value.title == title
    assert e.value.description == desc
    assert e.value.headers['foo'] == 'bar'


@pytest.mark.parametrize(
    'err, args',
    (
        (falcon.HTTPMethodNotAllowed, (['GET'],)),
        (falcon.HTTPRangeNotSatisfiable, (11,)),
        (falcon.HTTPInvalidHeader, ('foo', 'bar')),
        (falcon.HTTPMissingHeader, ('foo',)),
        (falcon.HTTPInvalidParam, ('foo', 'bar')),
        (falcon.HTTPMissingParam, ('foo',)),
    ),
)
def test_args_kw_only(err, args):
    with pytest.raises(TypeError, match='positional argument'):
        err(*args, 'bar')


@pytest.mark.parametrize(
    'err',
    [
        falcon.HTTPServiceUnavailable,
        falcon.HTTPTooManyRequests,
        falcon.HTTPContentTooLarge,
    ],
)
def test_with_retry_after(err):
    with pytest.raises(err) as e:
        raise err(retry_after='123')

    assert e.value.headers['Retry-After'] == '123'


@pytest.mark.parametrize(
    'err',
    [
        falcon.HTTPServiceUnavailable,
        falcon.HTTPTooManyRequests,
        falcon.HTTPContentTooLarge,
    ],
)
def test_with_retry_after_and_headers(err):
    with pytest.raises(err) as e:
        raise err(retry_after='123', headers={'foo': 'bar'})

    assert e.value.headers['Retry-After'] == '123'
    assert e.value.headers['foo'] == 'bar'


def test_http_error_repr():
    error = falcon.HTTPBadRequest()
    _repr = '<%s: %s>' % (error.__class__.__name__, error.status)
    assert error.__repr__() == _repr


@pytest.mark.parametrize(
    'err, args, title, desc',
    (
        (
            falcon.HTTPInvalidHeader,
            ('foo', 'bar'),
            'Invalid header value',
            'The value provided for the "bar" header is invalid. foo',
        ),
        (
            falcon.HTTPMissingHeader,
            ('foo',),
            'Missing header value',
            'The "foo" header is required.',
        ),
        (
            falcon.HTTPInvalidParam,
            ('foo', 'bar'),
            'Invalid parameter',
            'The "bar" parameter is invalid. foo',
        ),
        (
            falcon.HTTPMissingParam,
            ('foo',),
            'Missing parameter',
            'The "foo" parameter is required.',
        ),
    ),
)
def test_custom_400(err, args, title, desc):
    with pytest.raises(err) as e:
        raise err(*args)

    assert e.value.title == title
    assert e.value.description == desc


@pytest.mark.parametrize(
    'err, header_name, kw_name, args, res, kw_required',
    (
        (
            falcon.HTTPUnauthorized,
            'WWW-Authenticate',
            'challenges',
            ('a', 'b'),
            'a, b',
            False,
        ),
        (
            falcon.HTTPMethodNotAllowed,
            'Allow',
            'allowed_methods',
            ('a', 'b'),
            'a, b',
            True,
        ),
        (falcon.HTTPContentTooLarge, 'Retry-After', 'retry_after', 123, '123', False),
        (
            falcon.HTTPRangeNotSatisfiable,
            'Content-Range',
            'resource_length',
            123,
            'bytes */123',
            True,
        ),
        (falcon.HTTPTooManyRequests, 'Retry-After', 'retry_after', 123, '123', False),
        (
            falcon.HTTPServiceUnavailable,
            'Retry-After',
            'retry_after',
            123,
            '123',
            False,
        ),
    ),
)
class TestErrorsWithHeadersKW:
    def test_no_header(self, err, header_name, kw_name, args, res, kw_required):
        if not kw_required:
            value = err()

            if value.headers:
                assert header_name not in value.headers

    def test_other_header(self, err, header_name, kw_name, args, res, kw_required):
        headers = {'foo bar': 'baz'}
        kw = {kw_name: args}
        value = err(**kw, headers=headers)

        assert value.headers['foo bar'] == 'baz'
        assert header_name in value.headers
        assert value.headers[header_name] == res

    def test_override_header(self, err, header_name, kw_name, args, res, kw_required):
        headers = {'foo bar': 'baz', header_name: 'other'}
        kw = {kw_name: args}
        value = err(**kw, headers=headers)

        assert value.headers['foo bar'] == 'baz'
        assert header_name in value.headers
        assert value.headers[header_name] == res

    def test_other_header_list(self, err, header_name, kw_name, args, res, kw_required):
        headers = [('foo bar', 'baz')]
        kw = {kw_name: args}
        value = err(**kw, headers=headers)

        assert value.headers['foo bar'] == 'baz'
        assert header_name in value.headers
        assert isinstance(value.headers, dict)
        assert value.headers[header_name] == res

    def test_override_header_list(
        self, err, header_name, kw_name, args, res, kw_required
    ):
        headers = [('foo bar', 'baz'), (header_name, 'other')]
        kw = {kw_name: args}
        value = err(**kw, headers=headers)

        assert value.headers['foo bar'] == 'baz'
        assert header_name in value.headers
        assert isinstance(value.headers, dict)
        assert value.headers[header_name] == res


def test_http_payload_too_large_deprecation():
    with pytest.warns(match='HTTPContentTooLarge'):
        err = errors.HTTPPayloadTooLarge()
        assert err.title == '413 Content Too Large'



# Copyright 2016 by Rackspace Hosting, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""RFC extension for Sphinx.

This extensions adds hyperlinking for any RFC references that are
formatted like this::

    RFC 7231; Section 6.5.3
"""

import re

IETF_DOCS = 'https://datatracker.ietf.org/doc/html'
RFC_PATTERN = re.compile(r'RFC (\d{4}), Section ([\d\.]+)')


def _render_section(section_number, rfc_number):
    template = '`{0} <https://tools.ietf.org/html/rfc{1}#section-{0}>`_'
    return template.format(section_number, rfc_number)


def _process_line(line):
    m = RFC_PATTERN.search(line)
    if not m:
        return line

    rfc = m.group(1)
    section = m.group(2)

    template = (
        '`RFC {rfc}, Section {section} <{ietf_docs}/rfc{rfc}#section-{section}>`__'
    )

    rendered_text = template.format(rfc=rfc, section=section, ietf_docs=IETF_DOCS)

    return line[: m.start()] + rendered_text + line[m.end() :]


def _on_process_docstring(app, what, name, obj, options, lines):
    """Process the docstring for a given python object."""

    if what == 'module' and name == 'falcon':
        lines[:] = []
        return

    lines[:] = [_process_line(line) for line in lines]


def setup(app):
    app.connect('autodoc-process-docstring', _on_process_docstring)

    return {'parallel_read_safe': True}



import glob
import os
from os import path
import platform

from setuptools import setup

try:
    from Cython.Build import build_ext as _cy_build_ext
    from Cython.Distutils.extension import Extension as _cy_Extension

    HAS_CYTHON = True
except ImportError:
    _cy_build_ext = _cy_Extension = None
    HAS_CYTHON = False

DISABLE_EXTENSION = bool(os.environ.get('FALCON_DISABLE_CYTHON'))
IS_CPYTHON = platform.python_implementation() == 'CPython'

MYDIR = path.abspath(os.path.dirname(__file__))


if HAS_CYTHON and IS_CPYTHON and not DISABLE_EXTENSION:
    assert _cy_Extension is not None
    assert _cy_build_ext is not None

    def list_modules(dirname, pattern):
        filenames = glob.glob(path.join(dirname, pattern))

        module_names = []
        for name in filenames:
            module, ext = path.splitext(path.basename(name))
            if module != '__init__':
                module_names.append((module, ext))

        return module_names

    package_names = [
        'falcon.cyutil',
        # PERF(vytas): It seems that on recent (at the time of writing, 2025)
        #   CPython versions (especially 3.12+), cythonizing pure Python code
        #   is actually a de-optimization.
        # 'falcon',
        # 'falcon.media',
        # 'falcon.routing',
        # 'falcon.util',
    ]

    modules_to_exclude = [
        # NOTE(kgriffs): Cython does not handle dynamically-created async
        #   methods correctly.
        # NOTE(vytas,kgriffs): We have to also avoid cythonizing several
        #   other functions that might make it so that the framework
        #   can not recognize them as coroutine functions.
        #
        #   See also:
        #
        #       * https://github.com/cython/cython/issues/2273
        #       * https://bugs.python.org/issue38225
        #
        # NOTE(vytas): It is pointless to cythonize reader.py, since cythonized
        #   Falcon is using reader.pyx instead.
        'falcon.hooks',
        'falcon.inspect',
        'falcon.responders',
        'falcon.typing',
        'falcon._typing',
        'falcon.util.reader',
        'falcon.util.sync',
        'falcon.util.time',
    ]

    cython_package_names = ('falcon.cyutil',)
    # NOTE(vytas): Now that all our codebase is Python 3.8+, specify the
    #   Python 3 language level for Cython as well to avoid any surprises.
    cython_directives = {'language_level': '3', 'annotation_typing': False}

    ext_modules = [
        _cy_Extension(
            package + '.' + module,
            sources=[path.join(*(package.split('.') + [module + ext]))],
            cython_directives=cython_directives,
            optional=True,
        )
        for package in package_names
        for module, ext in list_modules(
            path.join(MYDIR, *package.split('.')),
            ('*.pyx' if package in cython_package_names else '*.py'),
        )
        if (package + '.' + module) not in modules_to_exclude
    ]

    cmdclass = {'build_ext': _cy_build_ext}
else:
    ext_modules = []
    cmdclass = {}


setup(cmdclass=cmdclass, ext_modules=ext_modules)



#!/usr/bin/env python

import argparse
import pathlib
import re
import sys

HERE = pathlib.Path(__file__).resolve().parent
FALCON_ROOT = HERE.parent
DIST = FALCON_ROOT / 'dist'

CHANGELOGS = FALCON_ROOT / 'docs' / 'changes'
RELEASE_STABLE_VERSION_PATTERN = re.compile(r'^(\d+)\.(\d+).(\d+)$')
RELEASE_META_PATTERN = re.compile(r'\.\.\s+falcon-release\:\s+([\d-]+)')


def check_dist(dist, git_ref):
    sdist = None
    versions = set()
    wheels = []

    if git_ref:
        git_ref = git_ref.split('/')[-1].lower()

    for path in dist.iterdir():
        if not path.is_file():
            continue

        if path.name.endswith('.tar.gz'):
            assert sdist is None, f'sdist already exists: {sdist}'
            sdist = path.name

        elif path.name.endswith('.whl'):
            wheels.append(path.name)

        else:
            sys.stderr.write(f'Unexpected file found in dist: {path.name}\n')
            sys.exit(1)

        package, _, _ = path.stem.partition('.tar')
        falcon, version, *_ = package.split('-')
        assert falcon == 'falcon', 'Unexpected package name: {path.name}'
        versions.add(version)

        if git_ref and version != git_ref:
            sys.stderr.write(
                f'Unexpected version: {path.name} ({version} != {git_ref})\n'
            )
            sys.exit(1)

    if not versions:
        sys.stderr.write('No artifacts collected!\n')
        sys.exit(1)
    if len(versions) > 1:
        sys.stderr.write(f'Multiple versions found: {tuple(versions)}!\n')
        sys.exit(1)
    version = versions.pop()

    wheel_list = '    None\n'
    if wheels:
        wheel_list = ''.join(f'    {wheel}\n' for wheel in sorted(wheels))

    release_date = None
    if RELEASE_STABLE_VERSION_PATTERN.match(version):
        changelog = CHANGELOGS / f'{version}.rst'
        if not changelog.is_file():
            sys.stderr.write(f'Changelog for stable release missing: {changelog}\n')
            sys.exit(1)
        if not (meta := RELEASE_META_PATTERN.search(changelog.read_text())):
            sys.stderr.write(f'falcon-release meta missing in: {changelog}\n')
            sys.exit(1)
        release_date = meta.group(1)

    print(f'[{dist}]\n')
    print(f'sdist found:\n    {sdist}\n')
    print(f'wheels found:\n{wheel_list}')
    print(f'version identified:\n    {version}\n')
    print(f'release date:\n    {release_date}\n')


def main():
    description = 'Check artifacts (sdist, wheels) inside dist dir.'

    parser = argparse.ArgumentParser(description=description)
    parser.add_argument(
        '-d',
        '--dist-dir',
        default=str(DIST),
        help='dist directory to check (default: %(default)s)',
    )
    parser.add_argument(
        '-r',
        '--git-ref',
        help='check version against git branch/tag ref (e.g. $GITHUB_REF)',
    )

    args = parser.parse_args()
    check_dist(pathlib.Path(args.dist_dir).resolve(), args.git_ref)


if __name__ == '__main__':
    main()



# Copied from the Requests library by Kenneth Reitz et al.
#
# Copyright 2013 Kenneth Reitz
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


"""Data structures.

This module provides additional data structures not found in the
standard library. These classes are hoisted into the `falcon` module
for convenience::

    import falcon

    things = falcon.CaseInsensitiveDict()
"""

from __future__ import annotations

from collections.abc import Mapping
from collections.abc import MutableMapping
from typing import (
    Any,
    Dict,
    ItemsView,
    Iterable,
    Iterator,
    KeysView,
    Optional,
    Tuple,
    TYPE_CHECKING,
    ValuesView,
)


# TODO(kgriffs): If we ever diverge from what is upstream in Requests,
# then we will need write tests and remove the "no cover" pragma.
class CaseInsensitiveDict(MutableMapping):  # pragma: no cover
    """A case-insensitive ``dict``-like object.

    Implements all methods and operations of
    ``collections.abc.MutableMapping`` as well as dict's `copy`. Also
    provides `lower_items`.

    All keys are expected to be strings. The structure remembers the
    case of the last key to be set, and ``iter(instance)``,
    ``keys()``, and ``items()``
    will contain case-sensitive keys. However, querying and contains
    testing is case insensitive:

        cid = CaseInsensitiveDict()
        cid['Accept'] = 'application/json'
        cid['aCCEPT'] == 'application/json'  # True
        list(cid) == ['Accept']  # True

    For example, ``headers['content-encoding']`` will return the
    value of a ``'Content-Encoding'`` response header, regardless
    of how the header name was originally stored.

    If the constructor, ``.update``, or equality comparison
    operations are given keys that have equal ``.lower()``s, the
    behavior is undefined.

    """

    def __init__(self, data: Optional[Iterable[Tuple[str, Any]]] = None, **kwargs: Any):
        self._store: Dict[str, Tuple[str, Any]] = dict()
        if data is None:
            data = {}
        self.update(data, **kwargs)

    def __setitem__(self, key: str, value: Any) -> None:
        # Use the lowercased key for lookups, but store the actual
        # key alongside the value.
        self._store[key.lower()] = (key, value)

    def __getitem__(self, key: str) -> Any:
        return self._store[key.lower()][1]

    def __delitem__(self, key: str) -> None:
        del self._store[key.lower()]

    def __iter__(self) -> Iterator[str]:
        return (casedkey for casedkey, mappedvalue in self._store.values())

    def __len__(self) -> int:
        return len(self._store)

    def lower_items(self) -> Iterator[Tuple[str, Any]]:
        """Like iteritems(), but with all lowercase keys."""
        return ((lowerkey, keyval[1]) for (lowerkey, keyval) in self._store.items())

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Mapping):
            other = CaseInsensitiveDict(other)
        else:
            return NotImplemented
        # Compare insensitively
        return dict(self.lower_items()) == dict(other.lower_items())

    # Copy is required
    def copy(self) -> CaseInsensitiveDict:
        return CaseInsensitiveDict(self._store.values())

    def __repr__(self) -> str:
        return '%s(%r)' % (self.__class__.__name__, dict(self.items()))


# NOTE(vytas): Although Context is effectively implementing the MutableMapping
#   interface, we choose not to subclass MutableMapping to stress the fact that
#   Context is, by design, a bare class, and the mapping interface may be
#   removed in a future Falcon release.
class Context:
    """Convenience class to hold contextual information in its attributes.

    This class is used as the default :class:`~.Request` and :class:`~Response`
    context type (see
    :attr:`Request.context_type <falcon.Request.context_type>` and
    :attr:`Response.context_type <falcon.Response.context_type>`,
    respectively).

    In Falcon versions prior to 2.0, the default context type was ``dict``. To
    ease the migration to attribute-based context object approach, this class
    also implements the mapping interface; that is, object attributes are
    linked to dictionary items, and vice versa. For instance:

    >>> context = falcon.Context()
    >>> context.cache_strategy = 'lru'
    >>> context.get('cache_strategy')
    'lru'
    >>> 'cache_strategy' in context
    True

    Although we have decided to maintain the mapping interface in the
    foreseeable future, new code should prefer the attribute-based approach, as
    it is more performant.

    What is more, if you continue to use the mapping interface
    (or mix-and-match), care needs to be taken not to overwrite :class:`dict`
    methods such as :meth:`~dict.items`, :meth:`~dict.values`, etc.
    """

    # NOTE(vytas): Define synthetic attr access methods (under TYPE_CHECKING)
    #   merely to let mypy know this is a namespace object.
    if TYPE_CHECKING:

        def __getattr__(self, name: str) -> Any: ...

        def __setattr__(self, name: str, value: Any) -> None: ...

        def __delattr__(self, name: str) -> None: ...

    def __contains__(self, key: str) -> bool:
        return self.__dict__.__contains__(key)

    def __getitem__(self, key: str) -> Optional[Any]:
        # PERF(vytas): On CPython, using this mapping interface (instead of a
        #   standard dict) to get, set and delete items incurs overhead
        #   approximately comparable to that of two function calls
        #   (per get/set/delete operation, that is).
        return self.__dict__.__getitem__(key)

    def __setitem__(self, key: str, value: Any) -> None:
        return self.__dict__.__setitem__(key, value)

    def __delitem__(self, key: str) -> None:
        self.__dict__.__delitem__(key)

    def __iter__(self) -> Iterator[str]:
        return self.__dict__.__iter__()

    def __len__(self) -> int:
        return self.__dict__.__len__()

    def __eq__(self, other: object) -> bool:
        if isinstance(other, type(self)):
            return self.__dict__.__eq__(other.__dict__)
        return self.__dict__.__eq__(other)

    def __ne__(self, other: object) -> bool:
        if isinstance(other, type(self)):
            return self.__dict__.__ne__(other.__dict__)
        return self.__dict__.__ne__(other)

    def __hash__(self) -> int:
        return hash(self.__dict__)

    def __repr__(self) -> str:
        return '{}({})'.format(type(self).__name__, self.__dict__.__repr__())

    def __str__(self) -> str:
        return '{}({})'.format(type(self).__name__, self.__dict__.__str__())

    def clear(self) -> None:
        return self.__dict__.clear()

    def copy(self) -> Context:
        ctx = type(self)()
        ctx.update(self.__dict__)
        return ctx

    def get(self, key: str, default: Optional[Any] = None) -> Optional[Any]:
        return self.__dict__.get(key, default)

    def items(self) -> ItemsView[str, Any]:
        return self.__dict__.items()

    def keys(self) -> KeysView[str]:
        return self.__dict__.keys()

    def pop(self, key: str, default: Optional[Any] = None) -> Optional[Any]:
        return self.__dict__.pop(key, default)

    def popitem(self) -> Tuple[str, Any]:
        return self.__dict__.popitem()

    def setdefault(
        self, key: str, default_value: Optional[Any] = None
    ) -> Optional[Any]:
        return self.__dict__.setdefault(key, default_value)

    def update(self, items: dict[str, Any]) -> None:
        self.__dict__.update(items)

    def values(self) -> ValuesView:
        return self.__dict__.values()


class ETag(str):
    """Convenience class to represent a parsed HTTP entity-tag.

    This class is simply a subclass of ``str`` with a few helper methods and
    an extra attribute to indicate whether the entity-tag is weak or strong. The
    value of the string is equivalent to what RFC 7232 calls an "opaque-tag",
    i.e. an entity-tag sans quotes and the weakness indicator.

    Note:

        Given that a weak entity-tag comparison can be performed by
        using the ``==`` operator (per the example below), only a
        :meth:`~.strong_compare` method is provided.

    Here is an example ``on_get()`` method that demonstrates how to use instances
    of this class::

        def on_get(self, req, resp):
            content_etag = self._get_content_etag()
            for etag in (req.if_none_match or []):
                if etag == '*' or etag == content_etag:
                    resp.status = falcon.HTTP_304
                    return

            # -- snip --

            resp.etag = content_etag
            resp.status = falcon.HTTP_200

    (See also: RFC 7232)
    """

    is_weak: bool = False
    """``True`` if the entity-tag is weak, otherwise ``False``."""

    def strong_compare(self, other: ETag) -> bool:
        """Perform a strong entity-tag comparison.

        Two entity-tags are equivalent if both are not weak and their
        opaque-tags match character-by-character.

        (See also: RFC 7232, Section 2.3.2)

        Arguments:
            other (ETag): The other :class:`~.ETag` to which you are comparing
            this one.

        Returns:
            bool: ``True`` if the two entity-tags match, otherwise ``False``.

        """

        return self == other and not (self.is_weak or other.is_weak)

    def dumps(self) -> str:
        """Serialize the ETag to a string suitable for use in a precondition header.

        (See also: RFC 7232, Section 2.3)

        Returns:
            str: An opaque quoted string, possibly prefixed by a weakness
            indicator ``W/``.
        """

        if self.is_weak:
            # PERF(kgriffs): Simple concatenation like this is slightly faster
            #   than %s string formatting.
            return 'W/"' + self + '"'

        return '"' + self + '"'

    @classmethod
    def loads(cls, etag_str: str) -> ETag:
        """Deserialize a single entity-tag string from a precondition header.

        Note:

            This method is meant to be used only for parsing a single
            entity-tag. It can not be used to parse a comma-separated list of
            values.

        (See also: RFC 7232, Section 2.3)

        Arguments:
            etag_str (str): An ASCII string representing a single entity-tag,
                as defined by RFC 7232.

        Returns:
            ETag: An instance of `~.ETag` representing the parsed entity-tag.

        """

        value = etag_str

        is_weak = False
        if value.startswith(('W/', 'w/')):
            is_weak = True
            value = value[2:]

        # NOTE(kgriffs): We allow for an unquoted entity-tag just in case,
        #   although it has been non-standard to do so since at least 1999
        #   with the advent of RFC 2616.
        if value[:1] == value[-1:] == '"':
            value = value[1:-1]

        t = cls(value)
        t.is_weak = is_weak

        return t



import http
import sys

import pytest

from falcon import status_codes


class TestStatusCodes:
    @pytest.mark.skipif(
        sys.version_info < (3, 13), reason='Outdated http statuses definitions'
    )
    @pytest.mark.parametrize('status', status_codes.__all__)
    def test_statuses_are_in_compliance_with_http_from_python313(self, status):
        status_code, message = self._status_code_and_message(status)
        if status_code >= 700:
            pytest.skip('Codes above 700 are not defined in http package')
        http_status = http.HTTPStatus(status_code)
        if status_code in [418, 422]:
            assert http_status.phrase != message
        else:
            assert http_status.phrase == message

    def _status_code_and_message(self, status: str):
        status = getattr(status_codes, status)
        value, message = status.split(' ', 1)
        return int(value), message



import falcon


class QuoteResource:
    def on_get(self, req: falcon.Request, resp: falcon.Response) -> None:
        """Handle GET requests."""
        resp.media = {
            'quote': "I've always been more interested in the future than in the past.",
            'author': 'Grace Hopper',
        }


app = falcon.App()
app.add_route('/quote', QuoteResource())



# Copyright 2021 by Kurt Griffiths
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Private args module extension for Sphinx.

This extension excludes underscore-prefixed args and kwargs from the function
signature.
"""


def _on_process_signature(app, what, name, obj, options, signature, return_annotation):
    if what in ('function', 'method') and signature and '_' in signature:
        filtered = []

        for token in signature[1:-1].split(','):
            token = token.strip()

            if not token.startswith('_'):
                filtered.append(token)

        signature = f'({", ".join(filtered)})'

    return signature, return_annotation


def setup(app):
    app.connect('autodoc-process-signature', _on_process_signature)

    return {'parallel_read_safe': True}



# This is a simple example of a WebSocket client that sends a message to the server.
# Since it's an example using the `websockets` library and it isn't using anything
# specific to Falcon, there are no tests. Coverage is skipped for this module.

import asyncio

import websockets


async def send_message():
    uri = 'ws://localhost:8000/reports'

    async with websockets.connect(uri) as websocket:
        # Send the authentication token
        await websocket.send('very secure token')

        while True:
            message = input('Name of the log (q to exit): ')
            if message.casefold() == 'q':
                break
            await websocket.send(message)
            response = await websocket.recv()
            print(response)


if __name__ == '__main__':
    asyncio.run(send_message())



# Copyright 2013 by Rackspace Hosting, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Response class."""

from __future__ import annotations

from datetime import datetime
from datetime import timezone
import functools
import mimetypes
from typing import (
    Any,
    ClassVar,
    Dict,
    Iterable,
    List,
    Mapping,
    NoReturn,
    Optional,
    overload,
    Tuple,
    Type,
    TYPE_CHECKING,
    Union,
)
import warnings

from falcon._typing import _UNSET
from falcon._typing import RangeSetHeader
from falcon._typing import UnsetOr
from falcon.constants import _DEFAULT_STATIC_MEDIA_TYPES
from falcon.constants import DEFAULT_MEDIA_TYPE
from falcon.errors import HeaderNotSupported
from falcon.media import Handlers
from falcon.response_helpers import _format_content_disposition
from falcon.response_helpers import _format_etag_header
from falcon.response_helpers import _format_header_value_list
from falcon.response_helpers import _format_range
from falcon.response_helpers import _header_property
from falcon.response_helpers import _is_ascii_encodable
from falcon.typing import Headers
from falcon.typing import ReadableIO
from falcon.util import dt_to_http
from falcon.util import http_cookies
from falcon.util import http_status_to_code
from falcon.util import structures
from falcon.util.deprecation import AttributeRemovedError
from falcon.util.deprecation import DeprecatedWarning
from falcon.util.uri import encode_check_escaped as uri_encode
from falcon.util.uri import encode_value_check_escaped as uri_encode_value

if TYPE_CHECKING:
    import http


_RESERVED_CROSSORIGIN_VALUES = frozenset({'anonymous', 'use-credentials'})
_RESERVED_SAMESITE_VALUES = frozenset({'lax', 'strict', 'none'})


class Response:
    """Represents an HTTP response to a client request.

    Note:
        ``Response`` is not meant to be instantiated directly by responders.

    Keyword Arguments:
        options (ResponseOptions): Set of global options passed from the App handler.
    """

    __slots__ = (
        'text',
        'context',
        'options',
        'status',
        'stream',
        '_cookies',
        '_data',
        '_extra_headers',
        '_headers',
        '_media',
        '_media_rendered',
        '__dict__',
    )

    _cookies: Optional[http_cookies.SimpleCookie]
    _data: Optional[bytes]
    _extra_headers: Optional[List[Tuple[str, str]]]
    _headers: Headers
    _media: Optional[Any]
    _media_rendered: UnsetOr[bytes]

    # Child classes may override this
    context_type: ClassVar[Type[structures.Context]] = structures.Context
    """Class variable that determines the factory or
    type to use for initializing the `context` attribute. By default,
    the framework will instantiate bare objects (instances of the bare
    :class:`falcon.Context` class). However, you may override this
    behavior by creating a custom child class of
    :class:`falcon.Response`, and then passing that new class to
    ``falcon.App()`` by way of the latter's `response_type` parameter.

    Note:
        When overriding `context_type` with a factory function (as
        opposed to a class), the function is called like a method of
        the current Response instance. Therefore the first argument is
        the Response instance itself (self).
    """

    # Attribute declaration
    complete: bool = False
    """Set to ``True`` from within a middleware method to signal to the framework that
    request processing should be short-circuited (see also
    :ref:`Middleware <middleware>`).
    """
    status: Union[str, int, http.HTTPStatus]
    """HTTP status code or line (e.g., ``'200 OK'``).

    This may be set to a member of :class:`http.HTTPStatus`, an HTTP status line
    string (e.g., ``'200 OK'``), or an ``int``.

    Note:
        The Falcon framework itself provides a number of constants for
        common status codes. They all start with the ``HTTP_`` prefix,
        as in: ``falcon.HTTP_204``. (See also: :ref:`status`.)
    """
    text: Optional[str]
    """String representing response content.

    Note:
        Falcon will encode the given text as UTF-8 in the response. If the content
        is already a byte string, use the :attr:`data` attribute instead (it's faster).
    """
    stream: Union[ReadableIO, Iterable[bytes], None]
    """Either a file-like object with a `read()` method that takes an optional size
    argument and returns a block of bytes, or an iterable object, representing response
    content, and yielding blocks as byte strings. Falcon will use *wsgi.file_wrapper*,
    if provided by the WSGI server, in order to efficiently serve file-like objects.

    Note:
        If the stream is set to an iterable object that requires
        resource cleanup, it can implement a close() method to do so.
        The close() method will be called upon completion of the request.
    """
    context: structures.Context
    """Empty object to hold any data (in its attributes) about the response which is
    specific to your app (e.g. session object).
    Falcon itself will not interact with this attribute after it has been initialized.

    Note:
        The preferred way to pass response-specific data, when using the
        default context type, is to set attributes directly on the
        `context` object. For example::

            resp.context.cache_strategy = 'lru'
    """
    options: ResponseOptions
    """Set of global options passed in from the App handler."""

    def __init__(self, options: Optional[ResponseOptions] = None) -> None:
        self.status = '200 OK'
        self._headers = {}

        # NOTE(kgriffs): Collection of additional headers as a list of raw
        #   tuples, to use in cases where we need more control over setting
        #   headers and duplicates are allowable or even necessary.
        #
        # PERF(kgriffs): Save some CPU cycles and a few bytes of RAM by
        #   only instantiating the list object later on IFF it is needed.
        self._extra_headers = None

        self.options = options if options is not None else ResponseOptions()

        # NOTE(tbug): will be set to a SimpleCookie object
        # when cookie is set via set_cookie
        self._cookies = None

        self.text = None
        self.stream = None
        self._data = None
        self._media = None
        self._media_rendered = _UNSET

        self.context = self.context_type()

    @property
    def status_code(self) -> int:
        """HTTP status code normalized from :attr:`status`.

        When a code is assigned to this property, :attr:`status` is updated,
        and vice-versa. The status code can be useful when needing to check
        in middleware for codes that fall into a certain class, e.g.::

            if resp.status_code >= 400:
                log.warning(f'returning error response: {resp.status_code}')
        """
        return http_status_to_code(self.status)

    @status_code.setter
    def status_code(self, value: int) -> None:
        self.status = value

    @property
    def body(self) -> NoReturn:
        raise AttributeRemovedError(
            'The body attribute is no longer supported. '
            'Please use the text attribute instead.'
        )

    @body.setter
    def body(self, value: str) -> NoReturn:
        raise AttributeRemovedError(
            'The body attribute is no longer supported. '
            'Please use the text attribute instead.'
        )

    @property
    def data(self) -> Optional[bytes]:
        """Byte string representing response content.

        Use this attribute in lieu of `text` when your content is
        already a byte string (of type ``bytes``). See also the note below.

        Warning:
            Always use the `text` attribute for text, or encode it
            first to ``bytes`` when using the `data` attribute, to
            ensure Unicode characters are properly encoded in the
            HTTP response.
        """
        return self._data

    @data.setter
    def data(self, value: Optional[bytes]) -> None:
        self._data = value

    @property
    def headers(self) -> Headers:
        """Copy of all headers set for the response, without cookies.

        Note that a new copy is created and returned each time this property is
        referenced.
        """
        return self._headers.copy()

    @property
    def media(self) -> Any:
        """A serializable object supported by the media handlers configured via
        :class:`falcon.RequestOptions`.

        Note:
            See also :ref:`media` for more information regarding media
            handling.
        """  # noqa D205
        return self._media

    @media.setter
    def media(self, value: Any) -> None:
        self._media = value
        self._media_rendered = _UNSET

    def render_body(self) -> Optional[bytes]:
        """Get the raw bytestring content for the response body.

        This method returns the raw data for the HTTP response body, taking
        into account the :attr:`~.text`, :attr:`~.data`, and :attr:`~.media`
        attributes.

        Note:
            This method ignores :attr:`~.stream`; the caller must check
            and handle that attribute directly.

        Returns:
            bytes: The UTF-8 encoded value of the `text` attribute, if
            set. Otherwise, the value of the `data` attribute if set, or
            finally the serialized value of the `media` attribute. If
            none of these attributes are set, ``None`` is returned.
        """
        data: Optional[bytes]
        text = self.text
        if text is None:
            data = self._data

            if data is None and self._media is not None:
                # NOTE(kgriffs): We use a special _UNSET singleton since
                #   None is ambiguous (the media handler might return None).
                if self._media_rendered is _UNSET:
                    if not self.content_type:
                        self.content_type = self.options.default_media_type

                    handler, _, _ = self.options.media_handlers._resolve(
                        self.content_type, self.options.default_media_type
                    )

                    self._media_rendered = handler.serialize(
                        self._media, self.content_type
                    )

                data = self._media_rendered
        else:
            try:
                # NOTE(kgriffs): Normally we expect text to be a string
                data = text.encode()
            except AttributeError:
                # NOTE(kgriffs): Assume it was a bytes object already
                data = text  # type: ignore[assignment]

        return data

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}: {self.status}>'

    def set_stream(
        self, stream: Union[ReadableIO, Iterable[bytes]], content_length: int
    ) -> None:
        """Set both `stream` and `content_length`.

        Although the :attr:`~falcon.Response.stream` and
        :attr:`~falcon.Response.content_length` properties may be set
        directly, using this method ensures
        :attr:`~falcon.Response.content_length` is not accidentally
        neglected when the length of the stream is known in advance. Using this
        method is also slightly more performant as compared to setting the
        properties individually.

        Note:
            If the stream length is unknown, you can set
            :attr:`~falcon.Response.stream` directly, and ignore
            :attr:`~falcon.Response.content_length`. In this case, the ASGI
            server may choose to use chunked encoding or one
            of the other strategies suggested by PEP-3333.

        Args:
            stream: A readable file-like object.
            content_length (int): Length of the stream, used for the
                Content-Length header in the response.
        """

        self.stream = stream

        # PERF(kgriffs): Set directly rather than incur the overhead of
        #   the self.content_length property.
        self._headers['content-length'] = str(content_length)

    def set_cookie(  # noqa: C901
        self,
        name: str,
        value: str,
        expires: Optional[datetime] = None,
        max_age: Optional[int] = None,
        domain: Optional[str] = None,
        path: Optional[str] = None,
        secure: Optional[bool] = None,
        http_only: bool = True,
        same_site: Optional[str] = None,
        partitioned: bool = False,
    ) -> None:
        """Set a response cookie.

        Note:
            This method can be called multiple times to add one or
            more cookies to the response.

        See Also:
            To learn more about setting cookies, see
            :ref:`Setting Cookies <setting-cookies>`. The parameters
            listed below correspond to those defined in `RFC 6265`_.

        Args:
            name (str): Cookie name
            value (str): Cookie value

        Keyword Args:
            expires (datetime): Specifies when the cookie should expire.
                By default, cookies expire when the user agent exits.

                (See also: RFC 6265, Section 4.1.2.1)
            max_age (int): Defines the lifetime of the cookie in
                seconds. By default, cookies expire when the user agent
                exits. If both `max_age` and `expires` are set, the
                latter is ignored by the user agent.

                Note:
                    Coercion to ``int`` is attempted if provided with
                    ``float`` or ``str``.

                (See also: RFC 6265, Section 4.1.2.2)

            domain (str): Restricts the cookie to a specific domain and
                any subdomains of that domain. By default, the user
                agent will return the cookie only to the origin server.
                When overriding this default behavior, the specified
                domain must include the origin server. Otherwise, the
                user agent will reject the cookie.

                Note:
                    Cookies do not provide isolation by port, so the domain
                    should not provide one. (See also: RFC 6265, Section 8.5)

                (See also: RFC 6265, Section 4.1.2.3)

            path (str): Scopes the cookie to the given path plus any
                subdirectories under that path (the "/" character is
                interpreted as a directory separator). If the cookie
                does not specify a path, the user agent defaults to the
                path component of the requested URI.

                Warning:
                    User agent interfaces do not always isolate
                    cookies by path, and so this should not be
                    considered an effective security measure.

                (See also: RFC 6265, Section 4.1.2.4)

            secure (bool): Direct the client to only return the cookie
                in subsequent requests if they are made over HTTPS
                (default: ``True``). This prevents attackers from
                reading sensitive cookie data.

                Note:
                    The default value for this argument is normally
                    ``True``, but can be modified by setting
                    :attr:`~.ResponseOptions.secure_cookies_by_default`
                    via :any:`App.resp_options`.

                Warning:
                    For the `secure` cookie attribute to be effective,
                    your application will need to enforce HTTPS.

                (See also: RFC 6265, Section 4.1.2.5)

            http_only (bool): The HttpOnly attribute limits the scope of the
                cookie to HTTP requests.  In particular, the attribute
                instructs the user agent to omit the cookie when providing
                access to cookies via "non-HTTP" APIs. This is intended to
                mitigate some forms of cross-site scripting. (default: ``True``)

                Note:
                    HttpOnly cookies are not visible to javascript scripts
                    in the browser. They are automatically sent to the server
                    on javascript ``XMLHttpRequest`` or ``Fetch`` requests.

                (See also: RFC 6265, Section 4.1.2.6)

            same_site (str): Helps protect against CSRF attacks by restricting
                when a cookie will be attached to the request by the user agent.
                When set to ``'Strict'``, the cookie will only be sent along
                with "same-site" requests.  If the value is ``'Lax'``, the
                cookie will be sent with same-site requests, and with
                "cross-site" top-level navigations.  If the value is ``'None'``,
                the cookie will be sent with same-site and cross-site requests.
                Finally, when this attribute is not set on the cookie, the
                attribute will be treated as if it had been set to ``'None'``.

                (See also: `Same-Site RFC Draft`_)

            partitioned (bool): Prevents cookies from being accessed from other
                subdomains. With partitioned enabled, a cookie set by
                https://3rd-party.example which is embedded inside
                https://site-a.example can no longer be accessed by
                https://site-b.example. While this attribute is not yet
                standardized, it is already used by Chrome.

                (See also: `CHIPS`_)

                .. versionadded:: 4.0

        Raises:
            KeyError: `name` is not a valid cookie name.
            ValueError: `value` is not a valid cookie value.

        .. _RFC 6265:
            http://tools.ietf.org/html/rfc6265

        .. _Same-Site RFC Draft:
            https://tools.ietf.org/html/draft-ietf-httpbis-rfc6265bis-03#section-4.1.2.7

        .. _CHIPS:
            https://developer.mozilla.org/en-US/docs/Web/Privacy/Privacy_sandbox/Partitioned_cookies

        """

        if not _is_ascii_encodable(name):
            raise KeyError('name is not ascii encodable')
        if not _is_ascii_encodable(value):
            raise ValueError('value is not ascii encodable')

        value = str(value)

        if self._cookies is None:
            self._cookies = http_cookies.SimpleCookie()

        try:
            self._cookies[name] = value
        except http_cookies.CookieError as e:  # pragma: no cover
            # NOTE(tbug): we raise a KeyError here, to avoid leaking
            # the CookieError to the user. SimpleCookie (well, BaseCookie)
            # only throws CookieError on issues with the cookie key
            raise KeyError(str(e))

        if expires:
            # set Expires on cookie. Format is Wdy, DD Mon YYYY HH:MM:SS GMT

            # NOTE(tbug): we never actually need to
            # know that GMT is named GMT when formatting cookies.
            # It is a function call less to just write "GMT" in the fmt string:
            fmt = '%a, %d %b %Y %H:%M:%S GMT'
            if expires.tzinfo is None:
                # naive
                self._cookies[name]['expires'] = expires.strftime(fmt)
            else:
                # aware
                gmt_expires = expires.astimezone(timezone.utc)
                self._cookies[name]['expires'] = gmt_expires.strftime(fmt)

        if max_age:
            # RFC 6265 section 5.2.2 says about the max-age value:
            #   "If the remainder of attribute-value contains a non-DIGIT
            #    character, ignore the cookie-av."
            # That is, RFC-compliant response parsers will ignore the max-age
            # attribute if the value contains a dot, as in floating point
            # numbers. Therefore, attempt to convert the value to an integer.
            self._cookies[name]['max-age'] = int(max_age)

        if domain:
            self._cookies[name]['domain'] = domain

        if path:
            self._cookies[name]['path'] = path

        is_secure = self.options.secure_cookies_by_default if secure is None else secure

        if is_secure:
            self._cookies[name]['secure'] = True

        if http_only:
            self._cookies[name]['httponly'] = http_only

        # PERF(kgriffs): Morsel.__setitem__() will lowercase this anyway,
        #   so we can just pass this in and when __setitem__() calls
        #   lower() it will be very slightly faster.
        if same_site:
            same_site = same_site.lower()

            if same_site not in _RESERVED_SAMESITE_VALUES:
                raise ValueError(
                    "same_site must be set to either 'lax', 'strict', or 'none'"
                )

            self._cookies[name]['samesite'] = same_site.capitalize()

        if partitioned:
            self._cookies[name]['partitioned'] = True

    def unset_cookie(
        self,
        name: str,
        domain: Optional[str] = None,
        path: Optional[str] = None,
        same_site: str = 'Lax',
        samesite: Optional[str] = None,
    ) -> None:
        """Unset a cookie in the response.

        Clears the contents of the cookie, and instructs the user
        agent to immediately expire its own copy of the cookie.

        Note:
            Modern browsers place restriction on cookies without the
            "same-site" cookie attribute set. To that end this attribute
            is set to ``'Lax'`` by this method.

            (See also: `Same-Site warnings`_)

        Warning:
            In order to successfully remove a cookie, both the
            path and the domain must match the values that were
            used when the cookie was created.

        Args:
            name (str): Cookie name

        Keyword Args:
            domain (str): Restricts the cookie to a specific domain and
                    any subdomains of that domain. By default, the user
                    agent will return the cookie only to the origin server.
                    When overriding this default behavior, the specified
                    domain must include the origin server. Otherwise, the
                    user agent will reject the cookie.

                    Note:
                        Cookies do not provide isolation by port, so the domain
                        should not provide one. (See also: RFC 6265, Section 8.5)

                    (See also: RFC 6265, Section 4.1.2.3)

            path (str): Scopes the cookie to the given path plus any
                subdirectories under that path (the "/" character is
                interpreted as a directory separator). If the cookie
                does not specify a path, the user agent defaults to the
                path component of the requested URI.

                Warning:
                    User agent interfaces do not always isolate
                    cookies by path, and so this should not be
                    considered an effective security measure.

                (See also: RFC 6265, Section 4.1.2.4)

            same_site (str): Allows to override the default 'Lax' same_site
                    setting for the unset cookie.

                    .. versionadded:: 4.1

            samesite (str): Deprecated. Use `same_site` instead.

                    .. deprecated:: 4.1
                       Please use the ``same_site`` parameter instead.

        .. _Same-Site warnings:
            https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Set-Cookie/SameSite#Fixing_common_warnings
        """  # noqa: E501
        if self._cookies is None:
            self._cookies = http_cookies.SimpleCookie()

        self._cookies[name] = ''

        # NOTE(Freezerburn): SimpleCookie apparently special cases the
        # expires attribute to automatically use strftime and set the
        # time as a delta from the current time. We use -1 here to
        # basically tell the browser to immediately expire the cookie,
        # thus removing it from future request objects.
        self._cookies[name]['expires'] = -1

        # Handle deprecated samesite parameter
        if samesite is not None:
            warnings.warn(
                'The "samesite" parameter is deprecated. '
                'Please use "same_site" instead.',
                DeprecatedWarning,
                stacklevel=2,
            )
            # Use the deprecated parameter value if same_site was not explicitly set
            same_site_value = samesite
        else:
            same_site_value = same_site

        # NOTE(CaselIT): Set SameSite to Lax to avoid setting invalid cookies.
        # See https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Set-Cookie/SameSite#Fixing_common_warnings  # noqa: E501
        self._cookies[name]['samesite'] = same_site_value

        if domain:
            self._cookies[name]['domain'] = domain

        if path:
            self._cookies[name]['path'] = path

    @overload
    def get_header(self, name: str, default: str) -> str: ...

    @overload
    def get_header(self, name: str, default: Optional[str] = ...) -> Optional[str]: ...

    def get_header(self, name: str, default: Optional[str] = None) -> Optional[str]:
        """Retrieve the raw string value for the given header.

        Normally, when a header has multiple values, they will be
        returned as a single, comma-delimited string. However, the
        Set-Cookie header does not support this format, and so
        attempting to retrieve it will raise an error.

        Args:
            name (str): Header name, case-insensitive. Must be of type ``str``
                or ``StringType``, and only character values 0x00 through 0xFF
                may be used on platforms that use wide characters.
        Keyword Args:
            default: Value to return if the header
                is not found (default ``None``).

        Raises:
            ValueError: The value of the 'Set-Cookie' header(s) was requested.

        Returns:
            str: The value of the specified header if set, or
            the default value if not set.
        """

        # NOTE(kgriffs): normalize name by lowercasing it
        name = name.lower()

        if name == 'set-cookie':
            raise HeaderNotSupported('Getting Set-Cookie is not currently supported.')

        return self._headers.get(name, default)

    def set_header(self, name: str, value: str) -> None:
        """Set a header for this response to a given value.

        Warning:
            Calling this method overwrites any values already set for this
            header. To append an additional value for this header, use
            :meth:`~.append_header` instead.

        Warning:
            This method cannot be used to set cookies; instead, use
            :meth:`~.append_header` or :meth:`~.set_cookie`.

        Args:
            name (str): Header name (case-insensitive). The name may contain
                only US-ASCII characters.
            value (str): Value for the header. As with the header's name, the
                value may contain only US-ASCII characters.

        Raises:
            ValueError: `name` cannot be ``'Set-Cookie'``.
        """

        # NOTE(kgriffs): uwsgi fails with a TypeError if any header
        # is not a str, so do the conversion here. It's actually
        # faster to not do an isinstance check. str() will encode
        # to US-ASCII.
        value = str(value)

        # NOTE(kgriffs): normalize name by lowercasing it
        name = name.lower()

        if name == 'set-cookie':
            raise HeaderNotSupported('This method cannot be used to set cookies')

        self._headers[name] = value

    def delete_header(self, name: str) -> None:
        """Delete a header that was previously set for this response.

        If the header was not previously set, nothing is done (no error is
        raised). Otherwise, all values set for the header will be removed
        from the response.

        Note that calling this method is equivalent to setting the
        corresponding header property (when said property is available) to
        ``None``. For example::

            resp.etag = None

        Warning:
            This method cannot be used with the Set-Cookie header. Instead,
            use :meth:`~.unset_cookie` to remove a cookie and ensure that the
            user agent expires its own copy of the data as well.

        Args:
            name (str): Header name (case-insensitive). The name may
                contain only US-ASCII characters.

        Raises:
            ValueError: `name` cannot be ``'Set-Cookie'``.
        """

        # NOTE(kgriffs): normalize name by lowercasing it
        name = name.lower()

        if name == 'set-cookie':
            raise HeaderNotSupported('This method cannot be used to remove cookies')

        self._headers.pop(name, None)

    def append_header(self, name: str, value: str) -> None:
        """Set or append a header for this response.

        If the header already exists, the new value will normally be appended
        to it, delimited by a comma. The notable exception to this rule is
        Set-Cookie, in which case a separate header line for each value will be
        included in the response.

        Note:
            While this method can be used to efficiently append raw
            Set-Cookie headers to the response, you may find
            :meth:`~.set_cookie` to be more convenient.

        Args:
            name (str): Header name (case-insensitive). The name may contain
                only US-ASCII characters.
            value (str): Value for the header. As with the header's name, the
                value may contain only US-ASCII characters.
        """

        # NOTE(kgriffs): uwsgi fails with a TypeError if any header
        # is not a str, so do the conversion here. It's actually
        # faster to not do an isinstance check. str() will encode
        # to US-ASCII.
        value = str(value)

        # NOTE(kgriffs): normalize name by lowercasing it
        name = name.lower()

        if name == 'set-cookie':
            if not self._extra_headers:
                self._extra_headers = [(name, value)]
            else:
                self._extra_headers.append((name, value))
        else:
            if name in self._headers:
                value = self._headers[name] + ', ' + value

            self._headers[name] = value

    def set_headers(
        self, headers: Union[Mapping[str, str], Iterable[Tuple[str, str]]]
    ) -> None:
        """Set several headers at once.

        This method can be used to set a collection of raw header names and
        values all at once.

        Warning:
            Calling this method overwrites any existing values for the given
            header. If a list containing multiple instances of the same header
            is provided, only the last value will be used. To add multiple
            values to the response for a given header, see
            :meth:`~.append_header`.

        Warning:
            This method cannot be used to set cookies; instead, use
            :meth:`~.append_header` or :meth:`~.set_cookie`.

        Args:
            headers (Iterable[[str, str]]): An iterable of ``[name, value]`` two-member
                iterables, or a dict-like object that implements an ``items()`` method.
                Both *name* and *value* must be of type ``str`` and
                contain only US-ASCII characters.

                Note:
                    Falcon can process an iterable of tuples slightly faster
                    than a dict.

        Raises:
            ValueError: `headers` was not a ``dict`` or ``list`` of ``tuple``
                         or ``Iterable[[str, str]]``.
        """

        header_items = getattr(headers, 'items', None)

        if callable(header_items):
            headers = header_items()

        # NOTE(kgriffs): We can't use dict.update because we have to
        # normalize the header names.
        _headers = self._headers

        for name, value in headers:  # type: ignore[misc]
            # NOTE(kgriffs): uwsgi fails with a TypeError if any header
            # is not a str, so do the conversion here. It's actually
            # faster to not do an isinstance check. str() will encode
            # to US-ASCII.
            value = str(value)

            name = name.lower()
            if name == 'set-cookie':
                raise HeaderNotSupported('This method cannot be used to set cookies')

            _headers[name] = value

    def append_link(
        self,
        target: str,
        rel: str,
        title: Optional[str] = None,
        title_star: Optional[Tuple[str, str]] = None,
        anchor: Optional[str] = None,
        hreflang: Optional[Union[str, Iterable[str]]] = None,
        type_hint: Optional[str] = None,
        crossorigin: Optional[str] = None,
        link_extension: Optional[Iterable[Tuple[str, str]]] = None,
    ) -> None:
        """Append a link header to the response.

        (See also: RFC 5988, Section 1)

        Note:
            Calling this method repeatedly will cause each link to be
            appended to the Link header value, separated by commas.

        Args:
            target (str): Target IRI for the resource identified by the
                link. Will be converted to a URI, if necessary, per
                RFC 3987, Section 3.1.
            rel (str): Relation type of the link, such as "next" or
                "bookmark".

                (See also:
                http://www.iana.org/assignments/link-relations/link-relations.xhtml)

        Keyword Args:
            title (str): Human-readable label for the destination of
                the link (default ``None``). If the title includes non-ASCII
                characters, you will need to use `title_star` instead, or
                provide both a US-ASCII version using `title` and a
                Unicode version using `title_star`.
            title_star (tuple[str, str]): Localized title describing the
                destination of the link (default ``None``). The value must be a
                two-member tuple in the form of (*language-tag*, *text*),
                where *language-tag* is a standard language identifier as
                defined in RFC 5646, Section 2.1, and *text* is a Unicode
                string.

                Note:
                    *language-tag* may be an empty string, in which case the
                    client will assume the language from the general context
                    of the current request.

                Note:
                    *text* will always be encoded as UTF-8.

            anchor (str): Override the context IRI with a different URI
                (default None). By default, the context IRI for the link is
                simply the IRI of the requested resource. The value
                provided may be a relative URI.
            hreflang (str or iterable): Either a single *language-tag*, or
                a ``list`` or ``tuple`` of such tags to provide a hint to the
                client as to the language of the result of following the link.
                A list of tags may be given in order to indicate to the
                client that the target resource is available in multiple
                languages.
            type_hint(str): Provides a hint as to the media type of the
                result of dereferencing the link (default ``None``). As noted
                in RFC 5988, this is only a hint and does not override the
                Content-Type header returned when the link is followed.
            crossorigin(str):  Determines how cross origin requests are handled.
                Can take values 'anonymous' or 'use-credentials' or None.
                (See:
                https://www.w3.org/TR/html50/infrastructure.html#cors-settings-attribute)
            link_extension: Provides additional custom attributes, as
                described in RFC 8288, Section 3.4.2; each member of the iterable
                must be a two-tuple in the form of (*param*, *value*).

        """

        # PERF(kgriffs): Heuristic to detect possibility of an extension
        # relation type, in which case it will be a URL that may contain
        # reserved characters. Otherwise, don't waste time running the
        # string through uri.encode
        #
        # Example values for rel:
        #
        #     "next"
        #     "http://example.com/ext-type"
        #     "https://example.com/ext-type"
        #     "alternate http://example.com/ext-type"
        #     "http://example.com/ext-type alternate"
        #
        if '//' in rel:
            if ' ' in rel:
                rel = '"' + ' '.join([uri_encode(r) for r in rel.split()]) + '"'
            else:
                rel = f'"{uri_encode(rel)}"'

        value = '<' + uri_encode(target) + '>; rel=' + rel

        if title is not None:
            value += f'; title="{title}"'

        if title_star is not None:
            value += f"; title*=UTF-8'{title_star[0]}'{uri_encode_value(title_star[1])}"

        if type_hint is not None:
            value += f'; type="{type_hint}"'

        if hreflang is not None:
            if isinstance(hreflang, str):
                value += f'; hreflang={hreflang}'
            else:
                value += '; '
                value += '; '.join(['hreflang=' + lang for lang in hreflang])

        if anchor is not None:
            value += f'; anchor="{uri_encode(anchor)}"'

        if crossorigin is not None:
            crossorigin = crossorigin.lower()
            if crossorigin not in _RESERVED_CROSSORIGIN_VALUES:
                raise ValueError(
                    "crossorigin must be set to either 'anonymous' or 'use-credentials'"
                )
            if crossorigin == 'anonymous':
                value += '; crossorigin'
            else:  # crossorigin == 'use-credentials'
                # PERF(vytas): the only remaining value is inlined.
                # Un-inline in case more values are supported in the future.
                value += '; crossorigin="use-credentials"'

        if link_extension is not None:
            value += '; '
            value += '; '.join([f'{p}={v}' for p, v in link_extension])

        _headers = self._headers
        if 'link' in _headers:
            _headers['link'] += f', {value}'
        else:
            _headers['link'] = value

    @property
    def add_link(self) -> NoReturn:
        raise AttributeRemovedError(
            'The add_link() method is no longer supported. '
            'Please use append_link() instead.'
        )

    cache_control: Union[str, Iterable[str], None] = _header_property(
        'Cache-Control',
        """Set the Cache-Control header.

        Used to set a list of cache directives to use as the value of the
        Cache-Control header. The list will be joined with ", " to produce
        the value for the header.
        """,
        _format_header_value_list,
    )
    """Set the Cache-Control header.

    Used to set a list of cache directives to use as the value of the
    Cache-Control header. The list will be joined with ", " to produce
    the value for the header.
    """

    content_location: Optional[str] = _header_property(
        'Content-Location',
        """Set the Content-Location header.

        This value will be URI encoded per RFC 3986. If the value that is
        being set is already URI encoded it should be decoded first or the
        header should be set manually using the set_header method.
        """,
        uri_encode,
    )
    """Set the Content-Location header.

    This value will be URI encoded per RFC 3986. If the value that is
    being set is already URI encoded it should be decoded first or the
    header should be set manually using the set_header method.
    """

    content_length: Union[str, int, None] = _header_property(
        'Content-Length',
        """Set the Content-Length header.

        This property can be used for responding to HEAD requests when you
        aren't actually providing the response body, or when streaming the
        response. If either the `text` property or the `data` property is set
        on the response, the framework will force Content-Length to be the
        length of the given text bytes. Therefore, it is only necessary to
        manually set the content length when those properties are not used.

        Note:
            In cases where the response content is a stream (readable
            file-like object), Falcon will not supply a Content-Length header
            to the server unless `content_length` is explicitly set.
            Consequently, the server may choose to use chunked encoding in this
            case.

        """,
    )
    """Set the Content-Length header.

    This property can be used for responding to HEAD requests when you
    aren't actually providing the response body, or when streaming the
    response. If either the `text` property or the `data` property is set
    on the response, the framework will force Content-Length to be the
    length of the given text bytes. Therefore, it is only necessary to
    manually set the content length when those properties are not used.

    Note:
        In cases where the response content is a stream (readable
        file-like object), Falcon will not supply a Content-Length header
        to the server unless `content_length` is explicitly set.
        Consequently, the server may choose to use chunked encoding in this
        case.

    """

    content_range: Union[str, RangeSetHeader, None] = _header_property(
        'Content-Range',
        """A tuple to use in constructing a value for the Content-Range header.

        The tuple has the form (*start*, *end*, *length*, [*unit*]), where *start* and
        *end* designate the range (inclusive), and *length* is the
        total length, or '\\*' if unknown. You may pass ``int``'s for
        these numbers (no need to convert to ``str`` beforehand). The optional value
        *unit* describes the range unit and defaults to 'bytes'

        Note:
            You only need to use the alternate form, 'bytes \\*/1234', for
            responses that use the status '416 Range Not Satisfiable'. In this
            case, raising ``falcon.HTTPRangeNotSatisfiable`` will do the right
            thing.

        (See also: RFC 7233, Section 4.2)
        """,
        _format_range,
    )
    """A tuple to use in constructing a value for the Content-Range header.

    The tuple has the form (*start*, *end*, *length*, [*unit*]), where *start* and
    *end* designate the range (inclusive), and *length* is the
    total length, or '\\*' if unknown. You may pass ``int``'s for
    these numbers (no need to convert to ``str`` beforehand). The optional value
    *unit* describes the range unit and defaults to 'bytes'

    Note:
        You only need to use the alternate form, 'bytes \\*/1234', for
        responses that use the status '416 Range Not Satisfiable'. In this
        case, raising ``falcon.HTTPRangeNotSatisfiable`` will do the right
        thing.

    (See also: RFC 7233, Section 4.2)
    """

    content_type: Optional[str] = _header_property(
        'Content-Type',
        """Sets the Content-Type header.

        The ``falcon`` module provides a number of constants for
        common media types, including ``falcon.MEDIA_JSON``,
        ``falcon.MEDIA_MSGPACK``, ``falcon.MEDIA_YAML``,
        ``falcon.MEDIA_XML``, ``falcon.MEDIA_HTML``,
        ``falcon.MEDIA_JS``, ``falcon.MEDIA_TEXT``,
        ``falcon.MEDIA_JPEG``, ``falcon.MEDIA_PNG``,
        and ``falcon.MEDIA_GIF``.
        """,
    )
    """Sets the Content-Type header.

    The ``falcon`` module provides a number of constants for
    common media types, including ``falcon.MEDIA_JSON``,
    ``falcon.MEDIA_MSGPACK``, ``falcon.MEDIA_YAML``,
    ``falcon.MEDIA_XML``, ``falcon.MEDIA_HTML``,
    ``falcon.MEDIA_JS``, ``falcon.MEDIA_TEXT``,
    ``falcon.MEDIA_JPEG``, ``falcon.MEDIA_PNG``,
    and ``falcon.MEDIA_GIF``.
    """

    downloadable_as: Optional[str] = _header_property(
        'Content-Disposition',
        """Set the Content-Disposition header using the given filename.

        The value will be used for the ``filename`` directive. For example,
        given ``'report.pdf'``, the Content-Disposition header would be set
        to: ``'attachment; filename="report.pdf"'``.

        As per `RFC 6266 <https://tools.ietf.org/html/rfc6266#appendix-D>`_
        recommendations, non-ASCII filenames will be encoded using the
        ``filename*`` directive, whereas ``filename`` will contain the US
        ASCII fallback.
        """,
        functools.partial(_format_content_disposition, disposition_type='attachment'),
    )
    """Set the Content-Disposition header using the given filename.

    The value will be used for the ``filename`` directive. For example,
    given ``'report.pdf'``, the Content-Disposition header would be set
    to: ``'attachment; filename="report.pdf"'``.

    As per `RFC 6266 <https://tools.ietf.org/html/rfc6266#appendix-D>`_
    recommendations, non-ASCII filenames will be encoded using the
    ``filename*`` directive, whereas ``filename`` will contain the US
    ASCII fallback.
    """

    viewable_as: Optional[str] = _header_property(
        'Content-Disposition',
        """Set an inline Content-Disposition header using the given filename.

        The value will be used for the ``filename`` directive. For example,
        given ``'report.pdf'``, the Content-Disposition header would be set
        to: ``'inline; filename="report.pdf"'``.

        As per `RFC 6266 <https://tools.ietf.org/html/rfc6266#appendix-D>`_
        recommendations, non-ASCII filenames will be encoded using the
        ``filename*`` directive, whereas ``filename`` will contain the US
        ASCII fallback.

        .. versionadded:: 3.1
        """,
        functools.partial(_format_content_disposition, disposition_type='inline'),
    )
    """Set an inline Content-Disposition header using the given filename.

    The value will be used for the ``filename`` directive. For example,
    given ``'report.pdf'``, the Content-Disposition header would be set
    to: ``'inline; filename="report.pdf"'``.

    As per `RFC 6266 <https://tools.ietf.org/html/rfc6266#appendix-D>`_
    recommendations, non-ASCII filenames will be encoded using the
    ``filename*`` directive, whereas ``filename`` will contain the US
    ASCII fallback.

    .. versionadded:: 3.1
    """

    etag: Optional[str] = _header_property(
        'ETag',
        """Set the ETag header.

        The ETag header will be wrapped with double quotes ``"value"`` in case
        the user didn't pass it.
        """,
        _format_etag_header,
    )
    """Set the ETag header.

    The ETag header will be wrapped with double quotes ``"value"`` in case
    the user didn't pass it.
    """

    expires: Union[str, datetime, None] = _header_property(
        'Expires',
        """Set the Expires header. Set to a ``datetime`` (UTC) instance.

        Note:
            Falcon will format the ``datetime`` as an HTTP date string.
        """,
        dt_to_http,
    )
    """Set the Expires header. Set to a ``datetime`` (UTC) instance.

    Note:
        Falcon will format the ``datetime`` as an HTTP date string.
    """

    last_modified: Union[str, datetime, None] = _header_property(
        'Last-Modified',
        """Set the Last-Modified header. Set to a ``datetime`` (UTC) instance.

        Note:
            Falcon will format the ``datetime`` as an HTTP date string.
        """,
        dt_to_http,
    )
    """Set the Last-Modified header. Set to a ``datetime`` (UTC) instance.

    Note:
        Falcon will format the ``datetime`` as an HTTP date string.
    """

    location: Optional[str] = _header_property(
        'Location',
        """Set the Location header.

        This value will be URI encoded per RFC 3986. If the value that is
        being set is already URI encoded it should be decoded first or the
        header should be set manually using the set_header method.
        """,
        uri_encode,
    )
    """Set the Location header.

    This value will be URI encoded per RFC 3986. If the value that is
    being set is already URI encoded it should be decoded first or the
    header should be set manually using the set_header method.
    """

    retry_after: Union[int, str, None] = _header_property(
        'Retry-After',
        """Set the Retry-After header.

        The expected value is an integral number of seconds to use as the
        value for the header. The HTTP-date syntax is not supported.
        """,
    )
    """Set the Retry-After header.

    The expected value is an integral number of seconds to use as the
    value for the header. The HTTP-date syntax is not supported.
    """

    vary: Union[str, Iterable[str], None] = _header_property(
        'Vary',
        """Value to use for the Vary header.

        Set this property to an iterable of header names. For a single
        asterisk or field value, simply pass a single-element ``list``
        or ``tuple``.

        The "Vary" header field in a response describes what parts of
        a request message, aside from the method, Host header field,
        and request target, might influence the origin server's
        process for selecting and representing this response.  The
        value consists of either a single asterisk ("*") or a list of
        header field names (case-insensitive).

        (See also: RFC 7231, Section 7.1.4)
        """,
        _format_header_value_list,
    )
    """Value to use for the Vary header.

    Set this property to an iterable of header names. For a single
    asterisk or field value, simply pass a single-element ``list``
    or ``tuple``.

    The "Vary" header field in a response describes what parts of
    a request message, aside from the method, Host header field,
    and request target, might influence the origin server's
    process for selecting and representing this response.  The
    value consists of either a single asterisk ("*") or a list of
    header field names (case-insensitive).

    (See also: RFC 7231, Section 7.1.4)
    """

    accept_ranges: Optional[str] = _header_property(
        'Accept-Ranges',
        """Set the Accept-Ranges header.

        The Accept-Ranges header field indicates to the client which
        range units are supported (e.g. "bytes") for the target
        resource.

        If range requests are not supported for the target resource,
        the header may be set to "none" to advise the client not to
        attempt any such requests.

        Note:
            "none" is the literal string, not Python's built-in ``None``
            type.

        """,
    )
    """Set the Accept-Ranges header.

    The Accept-Ranges header field indicates to the client which
    range units are supported (e.g. "bytes") for the target
    resource.

    If range requests are not supported for the target resource,
    the header may be set to "none" to advise the client not to
    attempt any such requests.

    Note:
        "none" is the literal string, not Python's built-in ``None``
        type.

    """

    def _set_media_type(self, media_type: Optional[str] = None) -> None:
        """Set a content-type; wrapper around set_header.

        Args:
            media_type: Media type to use for the Content-Type
                header.

        """

        # PERF(kgriffs): Using "in" like this is faster than dict.setdefault()
        #   in most cases, except on PyPy where it is only a fraction of a
        #   nanosecond slower. Last tested on Python versions 3.5-3.7.
        if media_type is not None and 'content-type' not in self._headers:
            self._headers['content-type'] = media_type

    def _wsgi_headers(self, media_type: Optional[str] = None) -> list[tuple[str, str]]:
        """Convert headers into the format expected by WSGI servers.

        Args:
            media_type: Default media type to use for the Content-Type
                header if the header was not set explicitly (default ``None``).

        """

        headers = self._headers
        # PERF(vytas): uglier inline version of Response._set_media_type
        if media_type is not None and 'content-type' not in headers:
            headers['content-type'] = media_type

        items = list(headers.items())

        if self._extra_headers:
            items += self._extra_headers

        # NOTE(kgriffs): It is important to append these after self._extra_headers
        #   in case the latter contains Set-Cookie headers that should be
        #   overridden by a call to unset_cookie().
        if self._cookies is not None:
            # PERF(tbug):
            # The below implementation is ~23% faster than
            # the alternative:
            #
            #     self._cookies.output().split("\\r\\n")
            #
            # Even without the .split("\\r\\n"), the below
            # is still ~17% faster, so don't use .output()
            items += [('set-cookie', c.OutputString()) for c in self._cookies.values()]
        return items


class ResponseOptions:
    """Defines a set of configurable response options.

    An instance of this class is exposed via :attr:`falcon.App.resp_options`
    and :attr:`falcon.asgi.App.resp_options` for configuring certain
    :class:`~.Response` behaviors.
    """

    secure_cookies_by_default: bool
    """Set to ``False`` in development environments to make the ``secure`` attribute
    for all cookies. (default ``True``).

    This can make testing easier by not requiring HTTPS. Note, however, that this
    setting can be overridden via :meth:`~.Response.set_cookie()`'s ``secure`` kwarg.
    """
    default_media_type: str
    """The default Internet media type (RFC 2046) to use when rendering a response,
    when the Content-Type header is not set explicitly.

    This value is normally set to the media type provided when a :class:`falcon.App`
    is initialized; however, if created independently, this will default to
    :attr:`falcon.DEFAULT_MEDIA_TYPE`.
    """
    media_handlers: Handlers
    """A dict-like object for configuring the media-types to handle.

    Default handlers are provided for the ``application/json``,
    ``application/x-www-form-urlencoded`` and ``multipart/form-data`` media types.
    """
    static_media_types: Dict[str, str]
    """A mapping of dot-prefixed file extensions to Internet media types (RFC 2046).

    Defaults to ``mimetypes.types_map`` after calling ``mimetypes.init()``.
    """
    xml_error_serialization: bool
    """Set to ``False`` to disable automatic inclusion of the XML handler
    in the :ref:`default error serializer <errors>` (default ``True``).

    Enabling this option does not make Falcon automatically render all error
    responses in XML, but it is used only in the case the client prefers
    (via the ``Accept`` request header) XML to JSON and other configured media
    handlers.

    Note:
        Falcon 5.0 will either change the default to ``False``, or remove the
        automatic XML error serialization altogether.

    Note:
        This option has no effect when a custom error serializer, set using
        :meth:`~falcon.App.set_error_serializer`, is in use.

    .. versionadded:: 4.0
    """

    __slots__ = (
        'secure_cookies_by_default',
        'default_media_type',
        'media_handlers',
        'static_media_types',
        'xml_error_serialization',
    )

    def __init__(self) -> None:
        self.secure_cookies_by_default = True
        self.default_media_type = DEFAULT_MEDIA_TYPE
        self.media_handlers = Handlers()
        self.xml_error_serialization = True

        if not mimetypes.inited:
            mimetypes.init()
        self.static_media_types = mimetypes.types_map.copy()
        self.static_media_types.update(_DEFAULT_STATIC_MEDIA_TYPES)



import asyncio
import time


async def _say_hi():
    print(f'[{time.time()}] Hi!')


async def app(scope, receive, send):
    await send(
        {
            'type': 'http.response.start',
            'status': 200,
            'headers': [
                [b'content-type', b'application/json'],
            ],
        }
    )
    await send(
        {
            'type': 'http.response.body',
            'body': f'[{time.time()}] Hello world!'.encode(),
        }
    )

    asyncio.create_task(_say_hi())



import falcon


def test_asgi():
    # TODO(vytas): consider removing this file completely.
    #   Its only purpose left is verifying that ASGI_SUPPORTED is still
    #   available for compatibility.
    assert falcon.ASGI_SUPPORTED



# Copyright 2020-2025 by Vytautas Liuolia.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def encode_items_to_latin1(dict data not None):
    cdef list result = []
    cdef unicode key
    cdef unicode value

    for key, value in data.items():
        result.append((key.encode('latin1'), value.encode('latin1')))

    return result



# Copyright 2021-2025 by Vytautas Liuolia.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Private type aliases used internally by Falcon.."""

from __future__ import annotations

from enum import auto
from enum import Enum
import http
from http.cookiejar import Cookie
import sys
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Iterable,
    List,
    Literal,
    Mapping,
    Optional,
    Pattern,
    Protocol,
    Tuple,
    TYPE_CHECKING,
    TypeVar,
    Union,
)

# NOTE(vytas): Mypy still struggles to handle a conditional import in the EAFP
#   fashion, so we branch on Py version instead (which it does understand).
if sys.version_info >= (3, 11):
    from wsgiref.types import StartResponse as StartResponse
    from wsgiref.types import WSGIEnvironment as WSGIEnvironment
else:
    WSGIEnvironment = Dict[str, Any]
    StartResponse = Callable[[str, List[Tuple[str, str]]], Callable[[bytes], None]]

if TYPE_CHECKING:
    from falcon.asgi import Request as AsgiRequest
    from falcon.asgi import Response as AsgiResponse
    from falcon.asgi import WebSocket
    from falcon.asgi_spec import AsgiEvent
    from falcon.asgi_spec import AsgiSendMsg
    from falcon.http_error import HTTPError
    from falcon.request import Request
    from falcon.response import Response


class _Unset(Enum):
    UNSET = auto()


_T = TypeVar('_T')
_UNSET = _Unset.UNSET
UnsetOr = Union[Literal[_Unset.UNSET], _T]

Link = Dict[str, str]
CookieArg = Mapping[str, Union[str, Cookie]]
# Error handlers
ErrorHandler = Callable[['Request', 'Response', Exception, Dict[str, Any]], None]


class AsgiErrorHandler(Protocol):
    async def __call__(
        self,
        req: AsgiRequest,
        resp: Optional[AsgiResponse],
        error: Exception,
        params: Dict[str, Any],
        *,
        ws: Optional[WebSocket] = ...,
    ) -> None: ...


# Error serializers
ErrorSerializer = Callable[['Request', 'Response', 'HTTPError'], None]

# Sinks
SinkPrefix = Union[str, Pattern[str]]


class SinkCallable(Protocol):
    def __call__(
        self, req: Request, resp: Response, **kwargs: Optional[str]
    ) -> None: ...


class AsgiSinkCallable(Protocol):
    async def __call__(
        self, req: AsgiRequest, resp: AsgiResponse, **kwargs: Optional[str]
    ) -> None: ...


HeaderMapping = Mapping[str, str]
HeaderIter = Iterable[Tuple[str, str]]
HeaderArg = Union[HeaderMapping, HeaderIter]
ResponseStatus = Union[http.HTTPStatus, str, int]
StoreArg = Optional[Dict[str, Any]]
Resource = object
RangeSetHeader = Union[Tuple[int, int, int], Tuple[int, int, int, str]]


# WSGI
class ResponderMethod(Protocol):
    def __call__(
        self,
        resource: Resource,
        req: Request,
        resp: Response,
        **kwargs: Any,
    ) -> None: ...


class ResponderCallable(Protocol):
    def __call__(self, req: Request, resp: Response, **kwargs: Any) -> None: ...


ProcessRequestMethod = Callable[['Request', 'Response'], None]
ProcessResourceMethod = Callable[
    ['Request', 'Response', Resource, Dict[str, Any]], None
]
ProcessResponseMethod = Callable[['Request', 'Response', Resource, bool], None]


# ASGI
class AsgiResponderMethod(Protocol):
    async def __call__(
        self,
        resource: Resource,
        req: AsgiRequest,
        resp: AsgiResponse,
        **kwargs: Any,
    ) -> None: ...


class AsgiResponderCallable(Protocol):
    async def __call__(
        self, req: AsgiRequest, resp: AsgiResponse, **kwargs: Any
    ) -> None: ...


class AsgiResponderWsCallable(Protocol):
    async def __call__(
        self, req: AsgiRequest, ws: WebSocket, **kwargs: Any
    ) -> None: ...


AsgiReceive = Callable[[], Awaitable['AsgiEvent']]
AsgiSend = Callable[['AsgiSendMsg'], Awaitable[None]]
AsgiProcessRequestMethod = Callable[['AsgiRequest', 'AsgiResponse'], Awaitable[None]]
AsgiProcessResourceMethod = Callable[
    ['AsgiRequest', 'AsgiResponse', Resource, Dict[str, Any]], Awaitable[None]
]
AsgiProcessResponseMethod = Callable[
    ['AsgiRequest', 'AsgiResponse', Resource, bool], Awaitable[None]
]
AsgiProcessRequestWsMethod = Callable[['AsgiRequest', 'WebSocket'], Awaitable[None]]
AsgiProcessResourceWsMethod = Callable[
    ['AsgiRequest', 'WebSocket', Resource, Dict[str, Any]], Awaitable[None]
]
ResponseCallbacks = Union[
    Tuple[Callable[[], None], Literal[False]],
    Tuple[Callable[[], Awaitable[None]], Literal[True]],
]


# Routing

MethodDict = Union[
    Dict[str, ResponderCallable],
    Dict[str, Union[AsgiResponderCallable, AsgiResponderWsCallable]],
]


class FindMethod(Protocol):
    def __call__(
        self, uri: str, req: Optional[Request]
    ) -> Optional[Tuple[object, MethodDict, Dict[str, Any], Optional[str]]]: ...


# Media
class SerializeSync(Protocol):
    def __call__(self, media: Any, content_type: Optional[str] = ...) -> bytes: ...


DeserializeSync = Callable[[bytes], Any]

Responder = Union[ResponderMethod, AsgiResponderMethod]


# WSGI middleware interface
class WsgiMiddlewareWithProcessRequest(Protocol):
    """WSGI Middleware with request handler."""

    def process_request(self, req: Request, resp: Response) -> None: ...


class WsgiMiddlewareWithProcessResource(Protocol):
    """WSGI Middleware with resource handler."""

    def process_resource(
        self,
        req: Request,
        resp: Response,
        resource: object,
        params: Dict[str, Any],
    ) -> None: ...


class WsgiMiddlewareWithProcessResponse(Protocol):
    """WSGI Middleware with response handler."""

    def process_response(
        self, req: Request, resp: Response, resource: object, req_succeeded: bool
    ) -> None: ...


# ASGI lifespan middleware interface
class AsgiMiddlewareWithProcessStartup(Protocol):
    """ASGI middleware with startup handler."""

    async def process_startup(
        self, scope: Mapping[str, Any], event: Mapping[str, Any]
    ) -> None: ...


class AsgiMiddlewareWithProcessShutdown(Protocol):
    """ASGI middleware with shutdown handler."""

    async def process_shutdown(
        self, scope: Mapping[str, Any], event: Mapping[str, Any]
    ) -> None: ...


# ASGI middleware interface
class AsgiMiddlewareWithProcessRequest(Protocol):
    """ASGI middleware with request handler."""

    async def process_request(self, req: AsgiRequest, resp: AsgiResponse) -> None: ...


class AsgiMiddlewareWithProcessResource(Protocol):
    """ASGI middleware with resource handler."""

    async def process_resource(
        self,
        req: AsgiRequest,
        resp: AsgiResponse,
        resource: object,
        params: Mapping[str, Any],
    ) -> None: ...


class AsgiMiddlewareWithProcessResponse(Protocol):
    """ASGI middleware with response handler."""

    async def process_response(
        self,
        req: AsgiRequest,
        resp: AsgiResponse,
        resource: object,
        req_succeeded: bool,
    ) -> None: ...


# ASGI WebSocket middleware
class AsgiMiddlewareWithProcessRequestWs(Protocol):
    """ASGI middleware with WebSocket request handler."""

    async def process_request_ws(self, req: AsgiRequest, ws: WebSocket) -> None: ...


class AsgiMiddlewareWithProcessResourceWs(Protocol):
    """ASGI middleware with WebSocket resource handler."""

    async def process_resource_ws(
        self,
        req: AsgiRequest,
        ws: WebSocket,
        resource: object,
        params: Mapping[str, Any],
    ) -> None: ...


# Universal middleware that provides async versions via the _async postfix
class UniversalMiddlewareWithProcessRequest(WsgiMiddlewareWithProcessRequest, Protocol):
    """WSGI/ASGI middleware with request handler."""

    async def process_request_async(
        self, req: AsgiRequest, resp: AsgiResponse
    ) -> None: ...


class UniversalMiddlewareWithProcessResource(
    WsgiMiddlewareWithProcessResource, Protocol
):
    """WSGI/ASGI middleware with resource handler."""

    async def process_resource_async(
        self,
        req: AsgiRequest,
        resp: AsgiResponse,
        resource: object,
        params: Mapping[str, Any],
    ) -> None: ...


class UniversalMiddlewareWithProcessResponse(
    WsgiMiddlewareWithProcessResponse, Protocol
):
    """WSGI/ASGI middleware with response handler."""

    async def process_response_async(
        self,
        req: AsgiRequest,
        resp: AsgiResponse,
        resource: object,
        req_succeeded: bool,
    ) -> None: ...


# NOTE(jkmnt): This typing is far from perfect due to the Python typing limitations,
# but better than nothing. Middleware conforming to any protocol of the union
# will pass the type check. Other protocols violations are not checked.
SyncMiddleware = Union[
    WsgiMiddlewareWithProcessRequest,
    WsgiMiddlewareWithProcessResource,
    WsgiMiddlewareWithProcessResponse,
]
"""Synchronous (WSGI) application middleware.

This type alias reflects the middleware interface for
components that can be used with a WSGI app.
"""

AsyncMiddleware = Union[
    AsgiMiddlewareWithProcessRequest,
    AsgiMiddlewareWithProcessResource,
    AsgiMiddlewareWithProcessResponse,
    # Lifespan middleware
    AsgiMiddlewareWithProcessStartup,
    AsgiMiddlewareWithProcessShutdown,
    # WebSocket middleware
    AsgiMiddlewareWithProcessRequestWs,
    AsgiMiddlewareWithProcessResourceWs,
    # Universal middleware with process_*_async methods
    UniversalMiddlewareWithProcessRequest,
    UniversalMiddlewareWithProcessResource,
    UniversalMiddlewareWithProcessResponse,
]
"""Asynchronous (ASGI) application middleware.

This type alias reflects the middleware interface for components that can be
used with an ASGI app.
"""



import os

import pytest

import falcon
import falcon.asgi
import falcon.request
import falcon.testing as testing

SIZE_1_KB = 1024


@pytest.fixture
def resource():
    return testing.SimpleTestResourceAsync()


@pytest.fixture
def client():
    app = falcon.asgi.App()
    return testing.TestClient(app)


class TestRequestBody:
    def test_empty_body(self, client, resource):
        client.app.add_route('/', resource)
        client.simulate_request(path='/', body='')
        stream = resource.captured_req.stream
        assert stream.tell() == 0

    def test_tiny_body(self, client, resource):
        client.app.add_route('/', resource)
        expected_body = '.'

        headers = {'capture-req-body-bytes': '1'}
        client.simulate_request(path='/', body=expected_body, headers=headers)
        stream = resource.captured_req.stream

        assert resource.captured_req_body == expected_body.encode('utf-8')
        assert stream.tell() == 1

    def test_tiny_body_overflow(self, client, resource):
        client.app.add_route('/', resource)
        expected_body = '.'
        expected_len = len(expected_body)

        # Read too many bytes; shouldn't block
        headers = {'capture-req-body-bytes': str(len(expected_body) + 1)}
        client.simulate_request(path='/', body=expected_body, headers=headers)
        stream = resource.captured_req.stream

        assert resource.captured_req_body == expected_body.encode('utf-8')
        assert stream.tell() == expected_len

    @pytest.mark.parametrize(
        'body_length, content_length',
        [
            (1, 0),
            (2, 1),
            (3, 2),
            (100, None),
            (100, 50),
            (8192, 50),
        ],
    )
    async def test_content_length_smaller_than_body(self, body_length, content_length):
        body_in = os.urandom(body_length)

        scope = testing.create_scope(content_length=content_length)
        req_event_emitter = testing.ASGIRequestEventEmitter(body=body_in)
        req_event_emitter._emit_empty_chunks = False
        first_event = await req_event_emitter.emit()
        req = falcon.asgi.Request(scope, req_event_emitter, first_event=first_event)

        body_out = await req.bounded_stream.read()
        assert body_out == body_in[:content_length]

    def test_read_body(self, client, resource):
        client.app.add_route('/', resource)
        expected_body = testing.rand_string(SIZE_1_KB // 2, SIZE_1_KB)
        expected_len = len(expected_body)

        headers = {
            'Content-Length': str(expected_len),
            'Capture-Req-Body-Bytes': '-1',
        }
        client.simulate_request(path='/', body=expected_body, headers=headers)

        content_len = resource.captured_req.get_header('content-length')
        assert content_len == str(expected_len)

        stream = resource.captured_req.stream

        assert resource.captured_req_body == expected_body.encode('utf-8')
        assert stream.tell() == expected_len

    def test_bounded_stream_alias(self):
        scope = testing.create_scope()
        req_event_emitter = testing.ASGIRequestEventEmitter(b'', disconnect_at=0)
        req = falcon.asgi.Request(scope, req_event_emitter)

        assert req.bounded_stream is req.stream

    def test_request_repr(self):
        scope = testing.create_scope()
        req_event_emitter = testing.ASGIRequestEventEmitter(b'', disconnect_at=0)
        req = falcon.asgi.Request(scope, req_event_emitter)

        _repr = '<%s: %s %r>' % (req.__class__.__name__, req.method, req.url)
        assert req.__repr__() == _repr



from threading import Barrier
from threading import Thread
from time import sleep
from unittest.mock import MagicMock

import pytest

from falcon.routing import compiled
from falcon.routing import CompiledRouter
from falcon.routing import CompiledRouterOptions


def test_find_src(monkeypatch):
    called = False
    find = CompiledRouter.find

    def mock(*args):
        nonlocal called
        called = True
        find(*args)

    monkeypatch.setattr(CompiledRouter, 'find', mock)
    router = CompiledRouter()

    assert router.finder_src is not None
    assert called


@pytest.fixture
def patch_add_route(monkeypatch):
    add_route = CompiledRouter.add_route
    mock = MagicMock(side_effect=lambda *a, **k: add_route(router, *a, **k))
    monkeypatch.setattr(CompiledRouter, 'add_route', mock)
    router = CompiledRouter()
    return mock, router


@pytest.mark.parametrize(
    'kwargs',
    (
        {},
        {'compile': False},
        {'compile': False, 'suffix': 'other'},
        {'suffix': 'other'},
    ),
)
def test_no_compile_kw(patch_add_route, kwargs):
    mock, router = patch_add_route

    res = MockResource()
    router.add_route('/foo', res, **kwargs)

    assert mock.call_count == 1
    mock.assert_has_calls(((('/foo', res), kwargs),))
    assert router._find == router._compile_and_find


def test_compile(patch_add_route):
    mock, router = patch_add_route

    res = MockResource()
    router.add_route('/foo', res, compile=True)
    assert mock.call_count == 1
    assert router._find != router._compile_and_find


def test_add_route_after_first_request():
    router = CompiledRouter()

    router.add_route('/foo', MockResource())
    assert router.find('/foo') is not None
    assert router._find != router._compile_and_find

    router.add_route('/bar', MockResource(), suffix='other')
    assert router._find == router._compile_and_find
    assert router.find('/bar') is not None
    assert router._find != router._compile_and_find


def test_multithread_compile(monkeypatch):
    def side_effect():
        sleep(0.05)
        return lambda *args: None

    mock = MagicMock(side_effect=side_effect)
    monkeypatch.setattr(CompiledRouter, '_compile', mock)

    router = CompiledRouter()
    mr = MockResource()

    router.add_route('/foo', mr)

    calls = 0
    num_threads = 3
    barrier = Barrier(num_threads)

    def find():
        nonlocal calls
        barrier.wait()
        assert router.find('/foo') is None
        calls += 1

    threads = [Thread(target=find) for i in range(num_threads)]
    for t in threads:
        t.start()

    for t in threads:
        t.join()

    assert calls == 3
    mock.assert_called_once_with()


class MockResource:
    def on_get(self, req, res):
        pass

    def on_get_other(self, req, res):
        pass


def test_cannot_replace_compiled():
    opt = CompiledRouterOptions()
    with pytest.raises(AttributeError, match='Cannot set'):
        opt.converters = {}
    with pytest.raises(AttributeError, match='object has no attribute'):
        opt.other = 123


def test_converter_not_subclass():
    class X:
        def convert(self, v):
            return v

    router = CompiledRouter()
    router.options.converters['x'] = X

    router.add_route('/foo/{bar:x}', MockResource())
    res = router.find('/foo/bar')
    assert res is not None
    assert res[2] == {'bar': 'bar'}
    assert router.find('/foo/bar/bar') is None


def test_base_classes():
    with pytest.raises(NotImplementedError):
        compiled._CxChild().src(42)



import falcon.asgi
import falcon.uri


class RawPathComponent:
    async def process_request(self, req, resp):
        raw_path = req.scope.get('raw_path')

        # NOTE: Decode the raw path from the raw_path bytestring, disallowing
        #   non-ASCII characters, assuming they are correctly percent-coded.
        if raw_path:
            req.path = raw_path.decode('ascii')


class URLResource:
    async def on_get(self, req, resp, url):
        # NOTE: url here is potentially percent-encoded.
        url = falcon.uri.decode(url)

        resp.media = {'url': url}

    async def on_get_status(self, req, resp, url):
        # NOTE: url here is potentially percent-encoded.
        url = falcon.uri.decode(url)

        resp.media = {'cached': True}


app = falcon.asgi.App(middleware=[RawPathComponent()])
app.add_route('/cache/{url}', URLResource())
app.add_route('/cache/{url}/status', URLResource(), suffix='status')



import asyncio
import datetime
import io

import aiofiles
import PIL.Image

import falcon


class Image:
    def __init__(self, config, image_id, size):
        self._config = config

        self.image_id = image_id
        self.size = size
        self.modified = datetime.datetime.now(datetime.timezone.utc)

    @property
    def path(self):
        return self._config.storage_path / self.image_id

    @property
    def uri(self):
        return f'/images/{self.image_id}.jpeg'

    def serialize(self):
        return {
            'id': self.image_id,
            'image': self.uri,
            'modified': falcon.dt_to_http(self.modified),
            'size': self.size,
            'thumbnails': self.thumbnails(),
        }

    def thumbnails(self):
        def reductions(size, min_size):
            width, height = size
            factor = 2
            while width // factor >= min_size and height // factor >= min_size:
                yield (width // factor, height // factor)
                factor *= 2

        return [
            f'/thumbnails/{self.image_id}/{width}x{height}.jpeg'
            for width, height in reductions(self.size, self._config.min_thumb_size)
        ]


class Store:
    def __init__(self, config):
        self._config = config
        self._images = {}

    def _load_from_bytes(self, data):
        return PIL.Image.open(io.BytesIO(data))

    def _convert(self, image):
        rgb_image = image.convert('RGB')

        converted = io.BytesIO()
        rgb_image.save(converted, 'JPEG')
        return converted.getvalue()

    def _resize(self, data, size):
        image = PIL.Image.open(io.BytesIO(data))
        image.thumbnail(size)

        resized = io.BytesIO()
        image.save(resized, 'JPEG')
        return resized.getvalue()

    def get(self, image_id):
        return self._images.get(image_id)

    def list_images(self):
        return sorted(self._images.values(), key=lambda item: item.modified)

    async def make_thumbnail(self, image, size):
        async with aiofiles.open(image.path, 'rb') as img_file:
            data = await img_file.read()

        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, self._resize, data, size)

    async def save(self, image_id, data):
        loop = asyncio.get_running_loop()
        image = await loop.run_in_executor(None, self._load_from_bytes, data)
        converted = await loop.run_in_executor(None, self._convert, image)

        path = self._config.storage_path / image_id
        async with aiofiles.open(path, 'wb') as output:
            await output.write(converted)

        stored = Image(self._config, image_id, image.size)
        self._images[image_id] = stored
        return stored



# misc test for 100% coverage

from unittest.mock import MagicMock

import pytest

from falcon.asgi import App
from falcon.http_error import HTTPError
from falcon.http_status import HTTPStatus


async def test_http_status_not_impl():
    app = App()
    with pytest.raises(NotImplementedError):
        await app._http_status_handler(MagicMock(), None, HTTPStatus(200), {}, None)


async def test_http_error_not_impl():
    app = App()
    with pytest.raises(NotImplementedError):
        await app._http_error_handler(MagicMock(), None, HTTPError(400), {}, None)


async def test_python_error_not_impl():
    app = App()
    with pytest.raises(NotImplementedError):
        await app._python_error_handler(
            MagicMock(), None, ValueError('error'), {}, None
        )



from datetime import datetime
import logging
import pathlib

import uvicorn

from falcon import WebSocketDisconnected
import falcon.asgi
from falcon.asgi import Request
from falcon.asgi import WebSocket

logger = logging.getLogger('ws-logger')
logger.setLevel('INFO')
logger.addHandler(logging.StreamHandler())


REPORTS = {
    'report1': {
        'title': 'Report 1',
        'content': 'This is the content of report 1',
    },
    'report2': {
        'title': 'Report 2',
        'content': 'This is the content of report 2',
    },
    'report3': {
        'title': 'Report 3',
        'content': 'This is the content of report 3',
    },
    'report4': {
        'title': 'Report 4',
        'content': 'This is the content of report 4',
    },
}

app = falcon.asgi.App()


class LoggerMiddleware:
    async def process_request_ws(self, req: Request, ws: WebSocket):
        # This will be called for the HTTP request that initiates the
        #   WebSocket handshake before routing.
        pass

    async def process_resource_ws(self, req: Request, ws: WebSocket, resource, params):
        # This will be called for the HTTP request that initiates the
        #   WebSocket handshake after routing (if a route matches the
        #   request).
        logger.info('WebSocket connection established on %r', req.path)


class AuthMiddleware:
    def __init__(self, protected_routes: list[str] | None = None):
        if protected_routes is None:
            protected_routes = []

        self.protected_routes = protected_routes

    async def process_request_ws(self, req: Request, ws: WebSocket):
        # Opening a connection so we can receive the token
        await ws.accept()

        # Check if the route is protected
        if req.path not in self.protected_routes:
            return

        token = await ws.receive_text()

        if token != 'very secure token':
            await ws.close(1008)
            return

        # Never log tokens in production
        logger.info('Client with token %r Authenticated', token)


class HelloWorldResource:
    async def on_get(self, req, resp):
        resp.media = {'hello': 'world'}


class EchoWebSocketResource:
    async def on_websocket(self, req: Request, ws: WebSocket):
        while True:
            try:
                message = await ws.receive_text()
                await ws.send_media(
                    {'message': message, 'date': datetime.now().isoformat()}
                )
            except WebSocketDisconnected:
                return


class ReportsResource:
    async def on_websocket(self, req: Request, ws: WebSocket):
        while True:
            try:
                query = await ws.receive_text()
                report = REPORTS.get(query, None)
                logger.info('selected report: %s', report)

                if report is None:
                    await ws.send_media({'error': 'report not found'})
                    continue

                await ws.send_media({'report': report['title']})
            except WebSocketDisconnected:
                return


app.add_route('/hello', HelloWorldResource())
app.add_route('/echo', EchoWebSocketResource())
app.add_route('/reports', ReportsResource())

app.add_middleware(LoggerMiddleware())
app.add_middleware(AuthMiddleware(['/reports']))

# usually a web server, like Nginx or Caddy, should serve static assets, but
# for the purpose of this example we use falcon.
static_path = pathlib.Path(__file__).parent / 'static'
app.add_static_route('/', static_path, fallback_filename='index.html')

if __name__ == '__main__':
    uvicorn.run(app, host='localhost', port=8000)  # pragma: no cover



import io

import pytest

from falcon.stream import BoundedStream
from falcon.util.deprecation import DeprecatedWarning


@pytest.fixture
def bounded_stream():
    return BoundedStream(io.BytesIO(), 1024)


def test_not_writable(bounded_stream):
    assert not bounded_stream.writable()

    with pytest.raises(IOError):
        bounded_stream.write(b'something something')


def test_exhausted():
    bs = BoundedStream(io.BytesIO(b'foobar'), 6)
    assert not bs.eof
    with pytest.warns(DeprecatedWarning, match='Use `eof` instead'):
        assert not bs.is_exhausted
    assert bs.read() == b'foobar'
    assert bs.eof
    with pytest.warns(DeprecatedWarning, match='Use `eof` instead'):
        assert bs.is_exhausted



from copy import copy

import pytest
from ws_tutorial.app import app
from ws_tutorial.app import AuthMiddleware

from falcon import errors
from falcon import testing


@pytest.mark.asyncio
async def test_websocket_echo():
    async with testing.ASGIConductor(app) as conn:
        async with conn.simulate_ws('/echo') as ws:
            await ws.send_text('Hello, World!')
            response = await ws.receive_json()

            assert response['message'] == 'Hello, World!'


@pytest.mark.asyncio
async def test_resetting_auth_middleware():
    local_app = copy(app)
    local_app._middleware = None
    local_app.add_middleware(AuthMiddleware())

    async with testing.ASGIConductor(local_app) as conn:
        async with conn.simulate_ws('/reports') as ws:
            with pytest.raises(errors.WebSocketDisconnected):
                await ws.send_text('report1')
                await ws.receive_json()


@pytest.mark.asyncio
async def test_websocket_reports():
    async with testing.ASGIConductor(app) as conn:
        async with conn.simulate_ws('/reports') as ws:
            await ws.send_text('very secure token')
            await ws.send_text('report1')
            response = await ws.receive_json()

            assert response['report'] == 'Report 1'


@pytest.mark.asyncio
async def test_websocket_report_not_found():
    async with testing.ASGIConductor(app) as conn:
        async with conn.simulate_ws('/reports') as ws:
            await ws.send_text('very secure token')
            await ws.send_text('report10')
            response = await ws.receive_json()

            assert response['error'] == 'report not found'


@pytest.mark.asyncio
async def test_websocket_not_authenticated():
    async with testing.ASGIConductor(app) as conn:
        async with conn.simulate_ws('/reports') as ws:
            with pytest.raises(errors.WebSocketDisconnected):
                await ws.send_text('report1')
                await ws.receive_json()



# Copyright 2013 by Rackspace Hosting, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Primary package for Falcon, the minimalist web API framework.

Falcon is a minimalist web API framework for building speedy web APIs and app
backends. The `falcon` package can be used to directly access most of
the framework's classes, functions, and variables::

    import falcon

    app = falcon.App()
"""

import logging as _logging

__all__ = (
    # API interface
    'API',
    'App',
    'after',
    'before',
    'BoundedStream',
    'CORSMiddleware',
    'HTTPError',
    'HTTPStatus',
    'HTTPFound',
    'HTTPMovedPermanently',
    'HTTPPermanentRedirect',
    'HTTPSeeOther',
    'HTTPTemporaryRedirect',
    'Forwarded',
    'Request',
    'RequestOptions',
    'Response',
    'ResponseOptions',
    # Public constants
    'HTTP_METHODS',
    'WEBDAV_METHODS',
    'COMBINED_METHODS',
    'DEFAULT_MEDIA_TYPE',
    'MEDIA_BMP',
    'MEDIA_GIF',
    'MEDIA_HTML',
    'MEDIA_JPEG',
    'MEDIA_JS',
    'MEDIA_JSON',
    'MEDIA_MSGPACK',
    'MEDIA_MULTIPART',
    'MEDIA_PNG',
    'MEDIA_TEXT',
    'MEDIA_URLENCODED',
    'MEDIA_XML',
    'MEDIA_YAML',
    'SINGLETON_HEADERS',
    'WebSocketPayloadType',
    # Utilities
    'async_to_sync',
    'BufferedReader',
    'CaseInsensitiveDict',
    'code_to_http_status',
    'Context',
    'create_task',
    'deprecated',
    'dt_to_http',
    'ETag',
    'get_argnames',
    'get_bound_method',
    'get_running_loop',
    'http_cookies',
    'http_date_to_dt',
    'http_now',
    'http_status_to_code',
    'IS_64_BITS',
    'is_python_func',
    'mediatypes',
    'misc',
    'parse_header',
    'reader',
    'runs_sync',
    'secure_filename',
    'structures',
    'sync',
    'sync_to_async',
    'time',
    'TimezoneGMT',
    'to_query_str',
    'uri',
    'wrap_sync_to_async',
    'wrap_sync_to_async_unsafe',
    # Error classes
    'CompatibilityError',
    'DelimiterError',
    'HeaderNotSupported',
    'HTTPBadGateway',
    'HTTPBadRequest',
    'HTTPConflict',
    'HTTPFailedDependency',
    'HTTPForbidden',
    'HTTPGatewayTimeout',
    'HTTPGone',
    'HTTPInsufficientStorage',
    'HTTPInternalServerError',
    'HTTPInvalidHeader',
    'HTTPInvalidParam',
    'HTTPLengthRequired',
    'HTTPLocked',
    'HTTPLoopDetected',
    'HTTPMethodNotAllowed',
    'HTTPMissingHeader',
    'HTTPMissingParam',
    'HTTPNetworkAuthenticationRequired',
    'HTTPNotAcceptable',
    'HTTPNotFound',
    'HTTPNotImplemented',
    'HTTPContentTooLarge',
    'HTTPPayloadTooLarge',
    'HTTPPreconditionFailed',
    'HTTPPreconditionRequired',
    'HTTPRangeNotSatisfiable',
    'HTTPRequestHeaderFieldsTooLarge',
    'HTTPRouteNotFound',
    'HTTPServiceUnavailable',
    'HTTPTooManyRequests',
    'HTTPUnauthorized',
    'HTTPUnavailableForLegalReasons',
    'HTTPUnprocessableEntity',
    'HTTPUnsupportedMediaType',
    'HTTPUriTooLong',
    'HTTPVersionNotSupported',
    'InvalidMediaRange',
    'InvalidMediaType',
    'MediaMalformedError',
    'MediaNotFoundError',
    'MediaValidationError',
    'MultipartParseError',
    'OperationNotAllowed',
    'PayloadTypeError',
    'UnsupportedError',
    'UnsupportedScopeError',
    'WebSocketDisconnected',
    'WebSocketHandlerNotFound',
    'WebSocketPathNotFound',
    'WebSocketServerError',
    # HTTP status codes
    'HTTP_100',
    'HTTP_101',
    'HTTP_102',
    'HTTP_103',
    'HTTP_200',
    'HTTP_201',
    'HTTP_202',
    'HTTP_203',
    'HTTP_204',
    'HTTP_205',
    'HTTP_206',
    'HTTP_207',
    'HTTP_208',
    'HTTP_226',
    'HTTP_300',
    'HTTP_301',
    'HTTP_302',
    'HTTP_303',
    'HTTP_304',
    'HTTP_305',
    'HTTP_307',
    'HTTP_308',
    'HTTP_400',
    'HTTP_401',
    'HTTP_402',
    'HTTP_403',
    'HTTP_404',
    'HTTP_405',
    'HTTP_406',
    'HTTP_407',
    'HTTP_408',
    'HTTP_409',
    'HTTP_410',
    'HTTP_411',
    'HTTP_412',
    'HTTP_413',
    'HTTP_414',
    'HTTP_415',
    'HTTP_416',
    'HTTP_417',
    'HTTP_418',
    'HTTP_421',
    'HTTP_422',
    'HTTP_423',
    'HTTP_424',
    'HTTP_425',
    'HTTP_426',
    'HTTP_428',
    'HTTP_429',
    'HTTP_431',
    'HTTP_451',
    'HTTP_500',
    'HTTP_501',
    'HTTP_502',
    'HTTP_503',
    'HTTP_504',
    'HTTP_505',
    'HTTP_506',
    'HTTP_507',
    'HTTP_508',
    'HTTP_510',
    'HTTP_511',
    'HTTP_701',
    'HTTP_702',
    'HTTP_703',
    'HTTP_710',
    'HTTP_711',
    'HTTP_712',
    'HTTP_719',
    'HTTP_720',
    'HTTP_721',
    'HTTP_722',
    'HTTP_723',
    'HTTP_724',
    'HTTP_725',
    'HTTP_726',
    'HTTP_727',
    'HTTP_740',
    'HTTP_741',
    'HTTP_742',
    'HTTP_743',
    'HTTP_744',
    'HTTP_745',
    'HTTP_748',
    'HTTP_749',
    'HTTP_750',
    'HTTP_753',
    'HTTP_754',
    'HTTP_755',
    'HTTP_759',
    'HTTP_771',
    'HTTP_772',
    'HTTP_773',
    'HTTP_774',
    'HTTP_776',
    'HTTP_777',
    'HTTP_778',
    'HTTP_779',
    'HTTP_780',
    'HTTP_781',
    'HTTP_782',
    'HTTP_783',
    'HTTP_784',
    'HTTP_785',
    'HTTP_786',
    'HTTP_791',
    'HTTP_792',
    'HTTP_797',
    'HTTP_799',
    'HTTP_ACCEPTED',
    'HTTP_ALREADY_REPORTED',
    'HTTP_BAD_GATEWAY',
    'HTTP_BAD_REQUEST',
    'HTTP_CONFLICT',
    'HTTP_CONTENT_TOO_LARGE',
    'HTTP_CONTINUE',
    'HTTP_CREATED',
    'HTTP_EARLY_HINTS',
    'HTTP_EXPECTATION_FAILED',
    'HTTP_FAILED_DEPENDENCY',
    'HTTP_FORBIDDEN',
    'HTTP_FOUND',
    'HTTP_GATEWAY_TIMEOUT',
    'HTTP_GONE',
    'HTTP_HTTP_VERSION_NOT_SUPPORTED',
    'HTTP_IM_A_TEAPOT',
    'HTTP_IM_USED',
    'HTTP_INSUFFICIENT_STORAGE',
    'HTTP_INTERNAL_SERVER_ERROR',
    'HTTP_LENGTH_REQUIRED',
    'HTTP_LOCKED',
    'HTTP_LOOP_DETECTED',
    'HTTP_METHOD_NOT_ALLOWED',
    'HTTP_MISDIRECTED_REQUEST',
    'HTTP_MOVED_PERMANENTLY',
    'HTTP_MULTIPLE_CHOICES',
    'HTTP_MULTI_STATUS',
    'HTTP_NETWORK_AUTHENTICATION_REQUIRED',
    'HTTP_NON_AUTHORITATIVE_INFORMATION',
    'HTTP_NOT_ACCEPTABLE',
    'HTTP_NOT_EXTENDED',
    'HTTP_NOT_FOUND',
    'HTTP_NOT_IMPLEMENTED',
    'HTTP_NOT_MODIFIED',
    'HTTP_NO_CONTENT',
    'HTTP_OK',
    'HTTP_PARTIAL_CONTENT',
    'HTTP_PAYMENT_REQUIRED',
    'HTTP_PERMANENT_REDIRECT',
    'HTTP_PRECONDITION_FAILED',
    'HTTP_PRECONDITION_REQUIRED',
    'HTTP_PROCESSING',
    'HTTP_PROXY_AUTHENTICATION_REQUIRED',
    'HTTP_REQUESTED_RANGE_NOT_SATISFIABLE',
    'HTTP_REQUEST_HEADER_FIELDS_TOO_LARGE',
    'HTTP_REQUEST_TIMEOUT',
    'HTTP_REQUEST_URI_TOO_LONG',
    'HTTP_RESET_CONTENT',
    'HTTP_SEE_OTHER',
    'HTTP_SERVICE_UNAVAILABLE',
    'HTTP_SWITCHING_PROTOCOLS',
    'HTTP_TEMPORARY_REDIRECT',
    'HTTP_TOO_EARLY',
    'HTTP_TOO_MANY_REQUESTS',
    'HTTP_UNAUTHORIZED',
    'HTTP_UNAVAILABLE_FOR_LEGAL_REASONS',
    'HTTP_UNPROCESSABLE_ENTITY',
    'HTTP_UNSUPPORTED_MEDIA_TYPE',
    'HTTP_UPGRADE_REQUIRED',
    'HTTP_USE_PROXY',
    'HTTP_VARIANT_ALSO_NEGOTIATES',
)

# NOTE(kgriffs,vytas): Hoist classes and functions into the falcon namespace.
#   Please explicitly list ALL exports.
from falcon.app import API
from falcon.app import App
from falcon.constants import ASGI_SUPPORTED  # NOQA: F401
from falcon.constants import COMBINED_METHODS
from falcon.constants import DEFAULT_MEDIA_TYPE
from falcon.constants import HTTP_METHODS
from falcon.constants import MEDIA_BMP
from falcon.constants import MEDIA_CSV
from falcon.constants import MEDIA_GIF
from falcon.constants import MEDIA_HTML
from falcon.constants import MEDIA_JPEG
from falcon.constants import MEDIA_JS
from falcon.constants import MEDIA_JSON
from falcon.constants import MEDIA_MSGPACK
from falcon.constants import MEDIA_MULTIPART
from falcon.constants import MEDIA_PARQUET
from falcon.constants import MEDIA_PNG
from falcon.constants import MEDIA_TEXT
from falcon.constants import MEDIA_URLENCODED
from falcon.constants import MEDIA_XML
from falcon.constants import MEDIA_YAML
from falcon.constants import PYTHON_VERSION  # NOQA: F401
from falcon.constants import SINGLETON_HEADERS
from falcon.constants import WEBDAV_METHODS
from falcon.constants import WebSocketPayloadType
from falcon.errors import CompatibilityError
from falcon.errors import DelimiterError
from falcon.errors import HeaderNotSupported
from falcon.errors import HTTPBadGateway
from falcon.errors import HTTPBadRequest
from falcon.errors import HTTPConflict
from falcon.errors import HTTPContentTooLarge
from falcon.errors import HTTPFailedDependency
from falcon.errors import HTTPForbidden
from falcon.errors import HTTPGatewayTimeout
from falcon.errors import HTTPGone
from falcon.errors import HTTPInsufficientStorage
from falcon.errors import HTTPInternalServerError
from falcon.errors import HTTPInvalidHeader
from falcon.errors import HTTPInvalidParam
from falcon.errors import HTTPLengthRequired
from falcon.errors import HTTPLocked
from falcon.errors import HTTPLoopDetected
from falcon.errors import HTTPMethodNotAllowed
from falcon.errors import HTTPMissingHeader
from falcon.errors import HTTPMissingParam
from falcon.errors import HTTPNetworkAuthenticationRequired
from falcon.errors import HTTPNotAcceptable
from falcon.errors import HTTPNotFound
from falcon.errors import HTTPNotImplemented
from falcon.errors import HTTPPayloadTooLarge
from falcon.errors import HTTPPreconditionFailed
from falcon.errors import HTTPPreconditionRequired
from falcon.errors import HTTPRangeNotSatisfiable
from falcon.errors import HTTPRequestHeaderFieldsTooLarge
from falcon.errors import HTTPRouteNotFound
from falcon.errors import HTTPServiceUnavailable
from falcon.errors import HTTPTooManyRequests
from falcon.errors import HTTPUnauthorized
from falcon.errors import HTTPUnavailableForLegalReasons
from falcon.errors import HTTPUnprocessableEntity
from falcon.errors import HTTPUnsupportedMediaType
from falcon.errors import HTTPUriTooLong
from falcon.errors import HTTPVersionNotSupported
from falcon.errors import InvalidMediaRange
from falcon.errors import InvalidMediaType
from falcon.errors import MediaMalformedError
from falcon.errors import MediaNotFoundError
from falcon.errors import MediaValidationError
from falcon.errors import MultipartParseError
from falcon.errors import OperationNotAllowed
from falcon.errors import PayloadTypeError
from falcon.errors import UnsupportedError
from falcon.errors import UnsupportedScopeError
from falcon.errors import WebSocketDisconnected
from falcon.errors import WebSocketHandlerNotFound
from falcon.errors import WebSocketPathNotFound
from falcon.errors import WebSocketServerError
from falcon.hooks import after
from falcon.hooks import before
from falcon.http_error import HTTPError
from falcon.http_status import HTTPStatus
from falcon.middleware import CORSMiddleware
from falcon.redirects import HTTPFound
from falcon.redirects import HTTPMovedPermanently
from falcon.redirects import HTTPPermanentRedirect
from falcon.redirects import HTTPSeeOther
from falcon.redirects import HTTPTemporaryRedirect
from falcon.request import Forwarded
from falcon.request import Request
from falcon.request import RequestOptions
from falcon.response import Response
from falcon.response import ResponseOptions

# Hoist HTTP status codes.
from falcon.status_codes import HTTP_100
from falcon.status_codes import HTTP_101
from falcon.status_codes import HTTP_102
from falcon.status_codes import HTTP_103
from falcon.status_codes import HTTP_200
from falcon.status_codes import HTTP_201
from falcon.status_codes import HTTP_202
from falcon.status_codes import HTTP_203
from falcon.status_codes import HTTP_204
from falcon.status_codes import HTTP_205
from falcon.status_codes import HTTP_206
from falcon.status_codes import HTTP_207
from falcon.status_codes import HTTP_208
from falcon.status_codes import HTTP_226
from falcon.status_codes import HTTP_300
from falcon.status_codes import HTTP_301
from falcon.status_codes import HTTP_302
from falcon.status_codes import HTTP_303
from falcon.status_codes import HTTP_304
from falcon.status_codes import HTTP_305
from falcon.status_codes import HTTP_307
from falcon.status_codes import HTTP_308
from falcon.status_codes import HTTP_400
from falcon.status_codes import HTTP_401
from falcon.status_codes import HTTP_402
from falcon.status_codes import HTTP_403
from falcon.status_codes import HTTP_404
from falcon.status_codes import HTTP_405
from falcon.status_codes import HTTP_406
from falcon.status_codes import HTTP_407
from falcon.status_codes import HTTP_408
from falcon.status_codes import HTTP_409
from falcon.status_codes import HTTP_410
from falcon.status_codes import HTTP_411
from falcon.status_codes import HTTP_412
from falcon.status_codes import HTTP_413
from falcon.status_codes import HTTP_414
from falcon.status_codes import HTTP_415
from falcon.status_codes import HTTP_416
from falcon.status_codes import HTTP_417
from falcon.status_codes import HTTP_418
from falcon.status_codes import HTTP_421
from falcon.status_codes import HTTP_422
from falcon.status_codes import HTTP_423
from falcon.status_codes import HTTP_424
from falcon.status_codes import HTTP_425
from falcon.status_codes import HTTP_426
from falcon.status_codes import HTTP_428
from falcon.status_codes import HTTP_429
from falcon.status_codes import HTTP_431
from falcon.status_codes import HTTP_451
from falcon.status_codes import HTTP_500
from falcon.status_codes import HTTP_501
from falcon.status_codes import HTTP_502
from falcon.status_codes import HTTP_503
from falcon.status_codes import HTTP_504
from falcon.status_codes import HTTP_505
from falcon.status_codes import HTTP_506
from falcon.status_codes import HTTP_507
from falcon.status_codes import HTTP_508
from falcon.status_codes import HTTP_510
from falcon.status_codes import HTTP_511
from falcon.status_codes import HTTP_701
from falcon.status_codes import HTTP_702
from falcon.status_codes import HTTP_703
from falcon.status_codes import HTTP_710
from falcon.status_codes import HTTP_711
from falcon.status_codes import HTTP_712
from falcon.status_codes import HTTP_719
from falcon.status_codes import HTTP_720
from falcon.status_codes import HTTP_721
from falcon.status_codes import HTTP_722
from falcon.status_codes import HTTP_723
from falcon.status_codes import HTTP_724
from falcon.status_codes import HTTP_725
from falcon.status_codes import HTTP_726
from falcon.status_codes import HTTP_727
from falcon.status_codes import HTTP_740
from falcon.status_codes import HTTP_741
from falcon.status_codes import HTTP_742
from falcon.status_codes import HTTP_743
from falcon.status_codes import HTTP_744
from falcon.status_codes import HTTP_745
from falcon.status_codes import HTTP_748
from falcon.status_codes import HTTP_749
from falcon.status_codes import HTTP_750
from falcon.status_codes import HTTP_753
from falcon.status_codes import HTTP_754
from falcon.status_codes import HTTP_755
from falcon.status_codes import HTTP_759
from falcon.status_codes import HTTP_771
from falcon.status_codes import HTTP_772
from falcon.status_codes import HTTP_773
from falcon.status_codes import HTTP_774
from falcon.status_codes import HTTP_776
from falcon.status_codes import HTTP_777
from falcon.status_codes import HTTP_778
from falcon.status_codes import HTTP_779
from falcon.status_codes import HTTP_780
from falcon.status_codes import HTTP_781
from falcon.status_codes import HTTP_782
from falcon.status_codes import HTTP_783
from falcon.status_codes import HTTP_784
from falcon.status_codes import HTTP_785
from falcon.status_codes import HTTP_786
from falcon.status_codes import HTTP_791
from falcon.status_codes import HTTP_792
from falcon.status_codes import HTTP_797
from falcon.status_codes import HTTP_799
from falcon.status_codes import HTTP_ACCEPTED
from falcon.status_codes import HTTP_ALREADY_REPORTED
from falcon.status_codes import HTTP_BAD_GATEWAY
from falcon.status_codes import HTTP_BAD_REQUEST
from falcon.status_codes import HTTP_CONFLICT
from falcon.status_codes import HTTP_CONTENT_TOO_LARGE
from falcon.status_codes import HTTP_CONTINUE
from falcon.status_codes import HTTP_CREATED
from falcon.status_codes import HTTP_EARLY_HINTS
from falcon.status_codes import HTTP_EXPECTATION_FAILED
from falcon.status_codes import HTTP_FAILED_DEPENDENCY
from falcon.status_codes import HTTP_FORBIDDEN
from falcon.status_codes import HTTP_FOUND
from falcon.status_codes import HTTP_GATEWAY_TIMEOUT
from falcon.status_codes import HTTP_GONE
from falcon.status_codes import HTTP_HTTP_VERSION_NOT_SUPPORTED
from falcon.status_codes import HTTP_IM_A_TEAPOT
from falcon.status_codes import HTTP_IM_USED
from falcon.status_codes import HTTP_INSUFFICIENT_STORAGE
from falcon.status_codes import HTTP_INTERNAL_SERVER_ERROR
from falcon.status_codes import HTTP_LENGTH_REQUIRED
from falcon.status_codes import HTTP_LOCKED
from falcon.status_codes import HTTP_LOOP_DETECTED
from falcon.status_codes import HTTP_METHOD_NOT_ALLOWED
from falcon.status_codes import HTTP_MISDIRECTED_REQUEST
from falcon.status_codes import HTTP_MOVED_PERMANENTLY
from falcon.status_codes import HTTP_MULTI_STATUS
from falcon.status_codes import HTTP_MULTIPLE_CHOICES
from falcon.status_codes import HTTP_NETWORK_AUTHENTICATION_REQUIRED
from falcon.status_codes import HTTP_NO_CONTENT
from falcon.status_codes import HTTP_NON_AUTHORITATIVE_INFORMATION
from falcon.status_codes import HTTP_NOT_ACCEPTABLE
from falcon.status_codes import HTTP_NOT_EXTENDED
from falcon.status_codes import HTTP_NOT_FOUND
from falcon.status_codes import HTTP_NOT_IMPLEMENTED
from falcon.status_codes import HTTP_NOT_MODIFIED
from falcon.status_codes import HTTP_OK
from falcon.status_codes import HTTP_PARTIAL_CONTENT
from falcon.status_codes import HTTP_PAYMENT_REQUIRED
from falcon.status_codes import HTTP_PERMANENT_REDIRECT
from falcon.status_codes import HTTP_PRECONDITION_FAILED
from falcon.status_codes import HTTP_PRECONDITION_REQUIRED
from falcon.status_codes import HTTP_PROCESSING
from falcon.status_codes import HTTP_PROXY_AUTHENTICATION_REQUIRED
from falcon.status_codes import HTTP_REQUEST_HEADER_FIELDS_TOO_LARGE
from falcon.status_codes import HTTP_REQUEST_TIMEOUT
from falcon.status_codes import HTTP_REQUEST_URI_TOO_LONG
from falcon.status_codes import HTTP_REQUESTED_RANGE_NOT_SATISFIABLE
from falcon.status_codes import HTTP_RESET_CONTENT
from falcon.status_codes import HTTP_SEE_OTHER
from falcon.status_codes import HTTP_SERVICE_UNAVAILABLE
from falcon.status_codes import HTTP_SWITCHING_PROTOCOLS
from falcon.status_codes import HTTP_TEMPORARY_REDIRECT
from falcon.status_codes import HTTP_TOO_EARLY
from falcon.status_codes import HTTP_TOO_MANY_REQUESTS
from falcon.status_codes import HTTP_UNAUTHORIZED
from falcon.status_codes import HTTP_UNAVAILABLE_FOR_LEGAL_REASONS
from falcon.status_codes import HTTP_UNPROCESSABLE_ENTITY
from falcon.status_codes import HTTP_UNSUPPORTED_MEDIA_TYPE
from falcon.status_codes import HTTP_UPGRADE_REQUIRED
from falcon.status_codes import HTTP_USE_PROXY
from falcon.status_codes import HTTP_VARIANT_ALSO_NEGOTIATES
from falcon.stream import BoundedStream

# NOTE(kgriffs): Ensure that "from falcon import uri" will import
#   the same front-door module as "import falcon.uri". This works by
#   priming the import cache with the one we want.
import falcon.uri  # NOQA: F401

# Hoist utilities.
from falcon.util import async_to_sync
from falcon.util import BufferedReader
from falcon.util import CaseInsensitiveDict
from falcon.util import code_to_http_status
from falcon.util import Context
from falcon.util import create_task

# NOTE(kgriffs): Hosting only because this was previously referenced
#   in the docs as falcon.deprecated
from falcon.util import deprecated
from falcon.util import dt_to_http
from falcon.util import ETag
from falcon.util import get_argnames
from falcon.util import get_bound_method
from falcon.util import get_running_loop
from falcon.util import http_cookies
from falcon.util import http_date_to_dt
from falcon.util import http_now
from falcon.util import http_status_to_code
from falcon.util import IS_64_BITS
from falcon.util import is_python_func
from falcon.util import mediatypes
from falcon.util import misc
from falcon.util import parse_header
from falcon.util import reader
from falcon.util import runs_sync
from falcon.util import secure_filename
from falcon.util import structures
from falcon.util import sync
from falcon.util import sync_to_async
from falcon.util import sys  # NOQA: F401
from falcon.util import time
from falcon.util import TimezoneGMT
from falcon.util import to_query_str
from falcon.util import uri
from falcon.util import wrap_sync_to_async
from falcon.util import wrap_sync_to_async_unsafe

# Package version
from falcon.version import __version__  # NOQA: F401

# NOTE(kgriffs): Only to be used internally on the rare occasion that we
#   need to log something that we can't communicate any other way.
_logger = _logging.getLogger('falcon')
_logger.addHandler(_logging.NullHandler())



import json

import falcon


class CustomJSONHandler(falcon.media.BaseHandler):
    MAX_INDENT_LEVEL = 8

    def deserialize(self, stream, content_type, content_length):
        data = stream.read()
        return json.loads(data.decode())

    def serialize(self, media, content_type):
        _, params = falcon.parse_header(content_type)
        indent = params.get('indent')
        if indent is not None:
            try:
                indent = int(indent)
                # NOTE: Impose a reasonable indentation level limit.
                if indent < 0 or indent > self.MAX_INDENT_LEVEL:
                    indent = None
            except ValueError:
                # TODO: Handle invalid params?
                indent = None

        result = json.dumps(media, indent=indent, sort_keys=bool(indent))
        return result.encode()



import random

import pecan
from pecan import expose, response, request


_body = pecan.x_test_body
_headers = pecan.x_test_headers


class TestController:
    def __init__(self, account_id):
        self.account_id = account_id

    @expose(content_type='text/plain')
    def test(self):
        user_agent = request.headers['User-Agent']  # NOQA
        limit = request.params.get('limit', '10')  # NOQA
        response.headers.update(_headers)

        return _body


class HelloController:
    @expose()
    def _lookup(self, account_id, *remainder):
        return TestController(account_id), remainder


class RootController:
    @expose(content_type='text/plain')
    def index(self):
        response.headers.update(_headers)
        return _body

    hello = HelloController()
